# Error Classes
# -------------

# No Errors
SUCCESS          = MPI_SUCCESS
ERR_LASTCODE     = MPI_ERR_LASTCODE
# ABI
ERR_ABI          = MPI_ERR_ABI
# Object Handles
ERR_TYPE         = MPI_ERR_TYPE
ERR_REQUEST      = MPI_ERR_REQUEST
ERR_OP           = MPI_ERR_OP
ERR_GROUP        = MPI_ERR_GROUP
ERR_INFO         = MPI_ERR_INFO
ERR_ERRHANDLER   = MPI_ERR_ERRHANDLER
ERR_SESSION      = MPI_ERR_SESSION
ERR_COMM         = MPI_ERR_COMM
ERR_WIN          = MPI_ERR_WIN
ERR_FILE         = MPI_ERR_FILE
# Communication Arguments
ERR_BUFFER       = MPI_ERR_BUFFER
ERR_COUNT        = MPI_ERR_COUNT
ERR_TAG          = MPI_ERR_TAG
ERR_RANK         = MPI_ERR_RANK
ERR_ROOT         = MPI_ERR_ROOT
ERR_TRUNCATE     = MPI_ERR_TRUNCATE
# Multiple Completion
ERR_IN_STATUS    = MPI_ERR_IN_STATUS
ERR_PENDING      = MPI_ERR_PENDING
# Topology Arguments
ERR_TOPOLOGY     = MPI_ERR_TOPOLOGY
ERR_DIMS         = MPI_ERR_DIMS
# Other Arguments
ERR_ARG          = MPI_ERR_ARG
# Other Errors
ERR_OTHER        = MPI_ERR_OTHER
ERR_UNKNOWN      = MPI_ERR_UNKNOWN
ERR_INTERN       = MPI_ERR_INTERN
# Attributes
ERR_KEYVAL       = MPI_ERR_KEYVAL
# Memory Allocation
ERR_NO_MEM       = MPI_ERR_NO_MEM
# Info Object
ERR_INFO_KEY     = MPI_ERR_INFO_KEY
ERR_INFO_VALUE   = MPI_ERR_INFO_VALUE
ERR_INFO_NOKEY   = MPI_ERR_INFO_NOKEY
# Dynamic Process Management
ERR_SPAWN        = MPI_ERR_SPAWN
ERR_PORT         = MPI_ERR_PORT
ERR_SERVICE      = MPI_ERR_SERVICE
ERR_NAME         = MPI_ERR_NAME
ERR_PROC_ABORTED = MPI_ERR_PROC_ABORTED
# One-Sided Communications
ERR_BASE         = MPI_ERR_BASE
ERR_SIZE         = MPI_ERR_SIZE
ERR_DISP         = MPI_ERR_DISP
ERR_ASSERT       = MPI_ERR_ASSERT
ERR_LOCKTYPE     = MPI_ERR_LOCKTYPE
ERR_RMA_CONFLICT = MPI_ERR_RMA_CONFLICT
ERR_RMA_SYNC     = MPI_ERR_RMA_SYNC
ERR_RMA_RANGE    = MPI_ERR_RMA_RANGE
ERR_RMA_ATTACH   = MPI_ERR_RMA_ATTACH
ERR_RMA_SHARED   = MPI_ERR_RMA_SHARED
ERR_RMA_FLAVOR   = MPI_ERR_RMA_FLAVOR
# Input/Output
ERR_BAD_FILE     = MPI_ERR_BAD_FILE
ERR_NO_SUCH_FILE = MPI_ERR_NO_SUCH_FILE
ERR_FILE_EXISTS  = MPI_ERR_FILE_EXISTS
ERR_FILE_IN_USE  = MPI_ERR_FILE_IN_USE
ERR_AMODE        = MPI_ERR_AMODE
ERR_ACCESS       = MPI_ERR_ACCESS
ERR_READ_ONLY    = MPI_ERR_READ_ONLY
ERR_NO_SPACE     = MPI_ERR_NO_SPACE
ERR_QUOTA        = MPI_ERR_QUOTA
ERR_NOT_SAME     = MPI_ERR_NOT_SAME
ERR_IO           = MPI_ERR_IO
ERR_UNSUPPORTED_OPERATION  = MPI_ERR_UNSUPPORTED_OPERATION
ERR_UNSUPPORTED_DATAREP    = MPI_ERR_UNSUPPORTED_DATAREP
ERR_CONVERSION             = MPI_ERR_CONVERSION
ERR_DUP_DATAREP            = MPI_ERR_DUP_DATAREP
ERR_VALUE_TOO_LARGE        = MPI_ERR_VALUE_TOO_LARGE
# Process Fault Tolerance
ERR_REVOKED             = MPI_ERR_REVOKED
ERR_PROC_FAILED         = MPI_ERR_PROC_FAILED
ERR_PROC_FAILED_PENDING = MPI_ERR_PROC_FAILED_PENDING


def Get_error_class(int errorcode: int) -> int:
    """
    Convert an *error code* into an *error class*.
    """
    cdef int errorclass = MPI_SUCCESS
    CHKERR( MPI_Error_class(errorcode, &errorclass) )
    return errorclass


def Get_error_string(int errorcode: int) -> str:
    """
    Return the *error string* for a given *error class* or *error code*.
    """
    cdef char string[MPI_MAX_ERROR_STRING+1]
    cdef int resultlen = 0
    CHKERR( MPI_Error_string(errorcode, string, &resultlen) )
    return tompistr(string, resultlen)


def Add_error_class() -> int:
    """
    Add an *error class* to the known error classes.
    """
    cdef int errorclass = MPI_SUCCESS
    CHKERR( MPI_Add_error_class(&errorclass) )
    return errorclass


def Remove_error_class(int errorclass: int) -> None:
    """
    Remove an *error class* from the known error classes.
    """
    CHKERR( MPI_Remove_error_class(errorclass) )


def Add_error_code(int errorclass: int) -> int:
    """
    Add an *error code* to an *error class*.
    """
    cdef int errorcode = MPI_SUCCESS
    CHKERR( MPI_Add_error_code(errorclass, &errorcode) )
    return errorcode


def Remove_error_code(int errorcode: int) -> None:
    """
    Remove an *error code* from the known error codes.
    """
    CHKERR( MPI_Remove_error_code(errorcode) )


def Add_error_string(int errorcode: int, string: str) -> None:
    """
    Associate an *error string* with an *error class* or *error code*.
    """
    cdef char *cstring = NULL
    string = asmpistr(string, &cstring)
    CHKERR( MPI_Add_error_string(errorcode, cstring) )


def Remove_error_string(int errorcode: int) -> None:
    """
    Remove *error string* association from *error class* or *error code*.
    """
    CHKERR( MPI_Remove_error_string(errorcode) )



import sys

try:
    import threading

    HAVE_THREADING = True
except ImportError:
    import dummy_threading as threading

    HAVE_THREADING = False

VERBOSE = False
# VERBOSE = True

import mpi4py.rc  # noqa: E402

mpi4py.rc.thread_level = "multiple"

import mpiunittest as unittest  # noqa: E402

from mpi4py import MPI  # noqa: E402


class TestMPIThreads(unittest.TestCase):
    #
    def testThreadLevels(self):
        levels = [
            MPI.THREAD_SINGLE,
            MPI.THREAD_FUNNELED,
            MPI.THREAD_SERIALIZED,
            MPI.THREAD_MULTIPLE,
        ]
        for i in range(len(levels) - 1):
            self.assertLess(levels[i], levels[i + 1])
        try:
            provided = MPI.Query_thread()
            self.assertIn(provided, levels)
        except NotImplementedError:
            self.skipTest("mpi-query_thread")

    def testIsThreadMain(self):
        try:
            flag = MPI.Is_thread_main()
        except NotImplementedError:
            self.skipTest("mpi-is_thread_main")
        name = threading.current_thread().name
        main = (name == "MainThread") or not HAVE_THREADING
        self.assertEqual(flag, main)
        if VERBOSE:

            def log(m):
                return sys.stderr.write(m + "\n")

            log(f"{name}: MPI.Is_thread_main() -> {flag}")

    def testIsThreadMainInThread(self):
        try:
            provided = MPI.Query_thread()
        except NotImplementedError:
            self.skipTest("mpi-query_thread")
        self.testIsThreadMain()
        T = [threading.Thread(target=self.testIsThreadMain) for _ in range(5)]
        if provided == MPI.THREAD_MULTIPLE:
            for t in T:
                t.start()
            for t in T:
                t.join()
        elif provided == MPI.THREAD_SERIALIZED:
            for t in T:
                t.start()
                t.join()
        else:
            self.skipTest("mpi-thread_level")


if __name__ == "__main__":
    unittest.main()



# Demonstrate how to work with Python GPU arrays using CUDA-aware MPI.
# We choose the CuPy library for simplicity, but any CUDA array which
# has the __cuda_array_interface__ attribute defined will work.
#
# Run this script using the following command:
# mpiexec -n 2 python use_cupy.py

import cupy

from mpi4py import MPI

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

# Allreduce
sendbuf = cupy.arange(10, dtype="i")
recvbuf = cupy.empty_like(sendbuf)
# always make sure the GPU buffer is ready before any MPI operation
cupy.cuda.get_current_stream().synchronize()
comm.Allreduce(sendbuf, recvbuf)
assert cupy.allclose(recvbuf, sendbuf * size)

# Bcast
if rank == 0:
    buf = cupy.arange(100, dtype=cupy.complex64)
else:
    buf = cupy.empty(100, dtype=cupy.complex64)
cupy.cuda.get_current_stream().synchronize()
comm.Bcast(buf)
assert cupy.allclose(buf, cupy.arange(100, dtype=cupy.complex64))

# Send-Recv
if rank == 0:
    buf = cupy.arange(20, dtype=cupy.float64)
    cupy.cuda.get_current_stream().synchronize()
    comm.Send(buf, dest=1, tag=88)
else:
    buf = cupy.empty(20, dtype=cupy.float64)
    cupy.cuda.get_current_stream().synchronize()
    comm.Recv(buf, source=0, tag=88)
    assert cupy.allclose(buf, cupy.arange(20, dtype=cupy.float64))



#!/usr/bin/env python
import os
import sys

from auditwheel.main import main

if "repair" in sys.argv:
    arch = os.uname().machine
    x, y = (2, 5) if arch == "x86_64" else (2, 17)
    plat = f"manylinux_{x}_{y}_{arch}"
    args = ["--plat", plat, "--only-plat"]
    for name in ("mpi",):
        args += ["--exclude", f"lib{name}.so.*"]
    sys.argv.extend(args)

sys.exit(main())



from mpi4py import rc

rc.finalize = False

from mpi4py import MPI  # noqa: E402

assert MPI.Is_initialized()
assert not MPI.Is_finalized()

MPI.Finalize()
assert MPI.Is_initialized()
assert MPI.Is_finalized()



import itertools

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


class BaseTestP2PBuf:
    #
    COMM = MPI.COMM_NULL

    def testSendrecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size + 1):
                    with self.subTest(s=s):
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + 1)
                        self.COMM.Sendrecv(
                            sbuf.as_mpi(), dest, 0, rbuf.as_mpi(), source, 0
                        )
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:-1]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        self.assertEqual(rbuf[-1], check)

    def testISendrecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        try:
            self.COMM.Isendrecv(
                bytearray(1),
                dest,
                0,
                bytearray(1),
                source,
                0,
            ).Wait()
        except NotImplementedError:
            if MPI.Get_version() >= (4, 0):
                raise
            raise unittest.SkipTest("mpi-isendrecv") from None
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size + 1):
                    with self.subTest(s=s):
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + 1)
                        self.COMM.Isendrecv(
                            sbuf.as_mpi(),
                            dest,
                            0,
                            rbuf.as_mpi(),
                            source,
                            0,
                        ).Wait()
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:-1]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        self.assertEqual(rbuf[-1], check)

    def testSendrecvReplace(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size + 1):
                    with self.subTest(s=s):
                        buf = array(rank, typecode, s)
                        self.COMM.Sendrecv_replace(
                            buf.as_mpi(), dest, 0, source, 0
                        )
                        check = arrayimpl.scalar(source)
                        for value in buf:
                            self.assertEqual(value, check)

    def testISendrecvReplace(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        try:
            self.COMM.Isendrecv_replace(
                bytearray(1), dest, 0, source, 0
            ).Wait()
        except NotImplementedError:
            if MPI.Get_version() >= (4, 0):
                raise
            raise unittest.SkipTest("mpi-isendrecv") from None
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                for s in range(size + 1):
                    with self.subTest(s=s):
                        buf = array(rank, typecode, s)
                        self.COMM.Isendrecv_replace(
                            buf.as_mpi(), dest, 0, source, 0
                        ).Wait()
                        check = arrayimpl.scalar(source)
                        for value in buf:
                            self.assertEqual(value, check)

    def testSendRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                for s in range(size + 1):
                    with self.subTest(s=s):
                        #
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s)
                        mem = array(
                            0, typecode, 2 * (s + MPI.BSEND_OVERHEAD)
                        ).as_raw()
                        if size == 1:
                            MPI.Attach_buffer(mem)
                            rbuf = sbuf
                            MPI.Detach_buffer()
                        elif rank == 0:
                            MPI.Attach_buffer(mem)
                            self.COMM.Ibsend(sbuf.as_mpi(), 1, 0).Wait()
                            self.COMM.Bsend(sbuf.as_mpi(), 1, 0)
                            MPI.Detach_buffer()
                            self.COMM.Send(sbuf.as_mpi(), 1, 0)
                            self.COMM.Ssend(sbuf.as_mpi(), 1, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                        elif rank == 1:
                            self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                            self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                            MPI.Attach_buffer(mem)
                            self.COMM.Ibsend(sbuf.as_mpi(), 0, 0).Wait()
                            self.COMM.Bsend(sbuf.as_mpi(), 0, 0)
                            MPI.Detach_buffer()
                            self.COMM.Send(sbuf.as_mpi(), 0, 0)
                            self.COMM.Ssend(sbuf.as_mpi(), 0, 0)
                        else:
                            rbuf = sbuf
                        check = arrayimpl.scalar(s)
                        for value in rbuf:
                            self.assertEqual(value, check)
                        #
                        rank = self.COMM.Get_rank()
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s)
                        rreq = self.COMM.Irecv(rbuf.as_mpi(), rank, 0)
                        self.COMM.Rsend(sbuf.as_mpi(), rank, 0)
                        rreq.Wait()
                        for value in rbuf:
                            self.assertEqual(value, check)
                        rbuf = array(-1, typecode, s)
                        rreq = self.COMM.Irecv(rbuf.as_mpi(), rank, 0)
                        self.COMM.Irsend(sbuf.as_mpi(), rank, 0).Wait()
                        rreq.Wait()
                        for value in rbuf:
                            self.assertEqual(value, check)

    def testProcNull(self):
        comm = self.COMM
        #
        comm.Sendrecv(None, MPI.PROC_NULL, 0, None, MPI.PROC_NULL, 0)
        comm.Sendrecv_replace(None, MPI.PROC_NULL, 0, MPI.PROC_NULL, 0)
        #
        comm.Send(None, MPI.PROC_NULL)
        comm.Isend(None, MPI.PROC_NULL).Wait()
        #
        comm.Ssend(None, MPI.PROC_NULL)
        comm.Issend(None, MPI.PROC_NULL).Wait()
        #
        buf = MPI.Alloc_mem(MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        comm.Bsend(None, MPI.PROC_NULL)
        comm.Ibsend(None, MPI.PROC_NULL).Wait()
        MPI.Detach_buffer()
        MPI.Free_mem(buf)
        #
        comm.Rsend(None, MPI.PROC_NULL)
        comm.Irsend(None, MPI.PROC_NULL).Wait()
        #
        comm.Recv(None, MPI.PROC_NULL)
        comm.Irecv(None, MPI.PROC_NULL).Wait()

    @unittest.skipMPI("mpich(<4.1.0)")
    def testProcNullISendrecv(self):
        try:
            self.COMM.Isendrecv(
                None,
                MPI.PROC_NULL,
                0,
                None,
                MPI.PROC_NULL,
                0,
            ).Wait()
            self.COMM.Isendrecv_replace(
                None,
                MPI.PROC_NULL,
                0,
                MPI.PROC_NULL,
                0,
            ).Wait()
        except NotImplementedError:
            if MPI.Get_version() >= (4, 0):
                raise
            raise unittest.SkipTest("mpi-isendrecv") from None

    @unittest.skipMPI("mpich(==3.4.1)")
    def testProcNullPersistent(self):
        comm = self.COMM
        #
        req = comm.Send_init(None, MPI.PROC_NULL)
        req.Start()
        req.Wait()
        req.Free()
        #
        req = comm.Ssend_init(None, MPI.PROC_NULL)
        req.Start()
        req.Wait()
        req.Free()
        #
        buf = MPI.Alloc_mem(MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        req = comm.Bsend_init(None, MPI.PROC_NULL)
        req.Start()
        req.Wait()
        req.Free()
        MPI.Detach_buffer()
        MPI.Free_mem(buf)
        #
        req = comm.Rsend_init(None, MPI.PROC_NULL)
        req.Start()
        req.Wait()
        req.Free()
        #
        req = comm.Recv_init(None, MPI.PROC_NULL)
        req.Start()
        req.Wait()
        req.Free()

    def testConstructor(self):
        preq = self.COMM.Send_init(b"", MPI.PROC_NULL, 0)
        dupe = MPI.Prequest(preq)
        self.assertIs(type(dupe), MPI.Prequest)
        self.assertEqual(dupe, preq)
        dupe = MPI.Prequest.fromhandle(preq.handle)
        self.assertIs(type(dupe), MPI.Prequest)
        self.assertEqual(dupe, preq)
        if preq.toint() != -1:
            dupe = MPI.Prequest.fromint(preq.toint())
            self.assertIs(type(dupe), MPI.Prequest)
            self.assertEqual(dupe, preq)
        if preq.py2f() != -1:
            dupe = MPI.Prequest.f2py(preq.py2f())
            self.assertIs(type(dupe), MPI.Prequest)
            self.assertEqual(dupe, preq)
        dupe = MPI.Request(preq)
        self.assertIs(type(dupe), MPI.Request)
        self.assertEqual(dupe, preq)
        with self.assertRaises(TypeError):
            dupe = MPI.Grequest(preq)
        preq.Start()
        preq.Wait()
        preq.Free()

    def testPersistent(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                for s, xs in itertools.product(range(size + 1), range(3)):
                    with self.subTest(s=s, xs=xs):
                        #
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + xs)
                        sendreq = self.COMM.Send_init(sbuf.as_mpi(), dest, 0)
                        recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                        sendreq.Start()
                        recvreq.Start()
                        sendreq.Wait()
                        recvreq.Wait()
                        self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                        sendreq.Free()
                        recvreq.Free()
                        self.assertEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertEqual(recvreq, MPI.REQUEST_NULL)
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:s]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        for value in rbuf[s:]:
                            self.assertEqual(value, check)
                        #
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + xs)
                        sendreq = self.COMM.Send_init(sbuf.as_mpi(), dest, 0)
                        recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                        reqlist = [sendreq, recvreq]
                        MPI.Prequest.Startall(reqlist)
                        index1 = MPI.Prequest.Waitany(reqlist)
                        self.assertIn(index1, [0, 1])
                        self.assertNotEqual(reqlist[index1], MPI.REQUEST_NULL)
                        index2 = MPI.Prequest.Waitany(reqlist)
                        self.assertIn(index2, [0, 1])
                        self.assertNotEqual(reqlist[index2], MPI.REQUEST_NULL)
                        self.assertNotEqual(index1, index2)
                        index3 = MPI.Prequest.Waitany(reqlist)
                        self.assertEqual(index3, MPI.UNDEFINED)
                        for preq in reqlist:
                            self.assertNotEqual(preq, MPI.REQUEST_NULL)
                            preq.Free()
                            self.assertEqual(preq, MPI.REQUEST_NULL)
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:s]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        for value in rbuf[s:]:
                            self.assertEqual(value, check)
                        #
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + xs)
                        sendreq = self.COMM.Ssend_init(sbuf.as_mpi(), dest, 0)
                        recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                        sendreq.Start()
                        recvreq.Start()
                        sendreq.Wait()
                        recvreq.Wait()
                        self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                        sendreq.Free()
                        recvreq.Free()
                        self.assertEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertEqual(recvreq, MPI.REQUEST_NULL)
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:s]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        for value in rbuf[s:]:
                            self.assertEqual(value, check)
                        #
                        mem = array(
                            0, typecode, s + MPI.BSEND_OVERHEAD
                        ).as_raw()
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + xs)
                        MPI.Attach_buffer(mem)
                        sendreq = self.COMM.Bsend_init(sbuf.as_mpi(), dest, 0)
                        recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                        sendreq.Start()
                        recvreq.Start()
                        sendreq.Wait()
                        recvreq.Wait()
                        MPI.Detach_buffer()
                        self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                        sendreq.Free()
                        recvreq.Free()
                        self.assertEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertEqual(recvreq, MPI.REQUEST_NULL)
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:s]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        for value in rbuf[s:]:
                            self.assertEqual(value, check)
                        #
                        rank = self.COMM.Get_rank()
                        sbuf = array(s, typecode, s)
                        rbuf = array(-1, typecode, s + xs)
                        recvreq = self.COMM.Recv_init(rbuf.as_mpi(), rank, 0)
                        sendreq = self.COMM.Rsend_init(sbuf.as_mpi(), rank, 0)
                        recvreq.Start()
                        sendreq.Start()
                        recvreq.Wait()
                        sendreq.Wait()
                        self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                        sendreq.Free()
                        recvreq.Free()
                        self.assertEqual(sendreq, MPI.REQUEST_NULL)
                        self.assertEqual(recvreq, MPI.REQUEST_NULL)
                        check = arrayimpl.scalar(s)
                        for value in rbuf[:s]:
                            self.assertEqual(value, check)
                        check = arrayimpl.scalar(-1)
                        for value in rbuf[s:]:
                            self.assertEqual(value, check)

    def testProbe(self):
        comm = self.COMM.Dup()
        try:
            request = comm.Issend([None, 0, MPI.BYTE], comm.rank, 123)
            self.assertTrue(request)
            status = MPI.Status()
            comm.Probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertEqual(status.source, comm.rank)
            self.assertEqual(status.tag, 123)
            self.assertTrue(request)
            flag = request.Test()
            self.assertTrue(request)
            self.assertFalse(flag)
            comm.Recv([None, 0, MPI.BYTE], comm.rank, 123)
            self.assertTrue(request)
            flag = False
            while not flag:
                flag = request.Test()
            self.assertFalse(request)
            self.assertTrue(flag)
        finally:
            comm.Free()

    @unittest.skipMPI("MPICH1")
    def testProbeCancel(self):
        comm = self.COMM.Dup()
        try:
            request = comm.Issend([None, 0, MPI.BYTE], comm.rank, 123)
            status = MPI.Status()
            comm.Probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertEqual(status.source, comm.rank)
            self.assertEqual(status.tag, 123)
            request.Cancel()
            self.assertTrue(request)
            status = MPI.Status()
            request.Get_status(status)
            cancelled = status.Is_cancelled()
            if not cancelled:
                comm.Recv([None, 0, MPI.BYTE], comm.rank, 123)
                request.Wait()
            else:
                request.Free()
        finally:
            comm.Free()

    def testIProbe(self):
        comm = self.COMM.Dup()
        try:
            f = comm.Iprobe()
            self.assertFalse(f)
            f = comm.Iprobe(MPI.ANY_SOURCE)
            self.assertFalse(f)
            f = comm.Iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG)
            self.assertFalse(f)
            status = MPI.Status()
            f = comm.Iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertFalse(f)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag, MPI.ANY_TAG)
            self.assertEqual(status.error, MPI.SUCCESS)
        finally:
            comm.Free()


class TestP2PBufSelf(BaseTestP2PBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestP2PBufWorld(BaseTestP2PBuf, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestP2PBufSelfDup(TestP2PBufSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


@unittest.skipMPI("openmpi(<1.4.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestP2PBufWorldDup(TestP2PBufWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI


class GReqCtx:
    #
    source = 3
    tag = 7
    completed = False

    cancel_called = False
    free_called = False

    def query(self, status):
        status.Set_source(self.source)
        status.Set_tag(self.tag)

    def free(self):
        self.free_called = True

    def cancel(self, completed):
        self.cancel_called = True
        if completed is not self.completed:
            raise MPI.Exception(MPI.ERR_PENDING)


@unittest.skipMPI("MPI(<2.0)")
@unittest.skipMPI("openmpi(==4.1.0)")
class TestGrequest(unittest.TestCase):
    #
    def testConstructor(self):
        ctx = GReqCtx()
        greq = MPI.Grequest.Start(ctx.query, ctx.free, ctx.cancel)
        dupe = MPI.Grequest(greq)
        self.assertIs(type(dupe), MPI.Grequest)
        self.assertEqual(dupe, greq)
        dupe = MPI.Grequest.fromhandle(greq.handle)
        self.assertIs(type(dupe), MPI.Grequest)
        self.assertEqual(dupe, greq)
        if greq.toint() != -1:
            dupe = MPI.Grequest.fromint(greq.toint())
            self.assertIs(type(dupe), MPI.Grequest)
            self.assertEqual(dupe, greq)
        if greq.py2f() != -1:
            dupe = MPI.Grequest.f2py(greq.py2f())
            self.assertIs(type(dupe), MPI.Grequest)
            self.assertEqual(dupe, greq)
        dupe = MPI.Request(greq)
        self.assertIs(type(dupe), MPI.Request)
        self.assertEqual(dupe, greq)
        with self.assertRaises(TypeError):
            dupe = MPI.Prequest(greq)
        greq.Cancel()
        greq.Complete()
        greq.Wait()

    @unittest.skipMPI("openmpi")  # TODO(dalcinl): open-mpi/ompi#11681
    def testExceptionHandling(self):
        ctx = GReqCtx()

        def raise_mpi(*_args):
            raise MPI.Exception(MPI.ERR_BUFFER)

        def raise_rte(*_args):
            raise ValueError(42)

        def check_exc(exception, is_mpi, stderr):
            output = stderr.getvalue()
            header = "Traceback (most recent call last):\n"
            if is_mpi:
                chkcode = MPI.ERR_BUFFER
                excname = MPI.Exception.__name__
            else:
                chkcode = MPI.ERR_OTHER
                excname = ValueError.__name__
            ierr = exception.Get_error_class()
            self.assertEqual(ierr, chkcode)
            self.assertTrue(output.startswith(header))
            self.assertIn(excname, output)

        for raise_fn, is_mpi in (
            (raise_mpi, True),
            (raise_rte, False),
        ):
            greq = MPI.Grequest.Start(raise_fn, ctx.free, ctx.cancel)
            greq.Complete()
            with self.assertRaises(MPI.Exception) as exc_cm:
                with testutil.capture_stderr() as stderr:
                    greq.Wait()
            if greq:
                greq.Free()
            check_exc(exc_cm.exception, is_mpi, stderr)
            #
            greq = MPI.Grequest.Start(ctx.query, raise_fn, ctx.cancel)
            greq.Complete()
            with self.assertRaises(MPI.Exception) as exc_cm:
                with testutil.capture_stderr() as stderr:
                    greq.Wait()
            if greq:
                greq.Free()
            check_exc(exc_cm.exception, is_mpi, stderr)
            #
            greq = MPI.Grequest.Start(ctx.query, ctx.free, raise_fn)
            with self.assertRaises(MPI.Exception) as exc_cm:
                with testutil.capture_stderr() as stderr:
                    greq.Cancel()
            greq.Complete()
            greq.Wait()
            if greq:
                greq.Free()
            check_exc(exc_cm.exception, is_mpi, stderr)

    def testAll(self):
        ctx = GReqCtx()
        greq = MPI.Grequest.Start(ctx.query, ctx.free, ctx.cancel)
        self.assertFalse(greq.Test())
        self.assertFalse(ctx.free_called)
        greq.Cancel()
        self.assertTrue(ctx.cancel_called)
        ctx.cancel_called = False
        greq.Complete()
        ctx.completed = True
        greq.Cancel()
        self.assertTrue(ctx.cancel_called)
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), ctx.source)
        self.assertEqual(status.Get_tag(), ctx.tag)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        greq.Wait()
        self.assertTrue(ctx.free_called)

    def testAll1(self):
        ctx = GReqCtx()
        greq = MPI.Grequest.Start(ctx.query, None, None)
        self.assertFalse(greq.Test())
        greq.Cancel()
        greq.Complete()
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), ctx.source)
        self.assertEqual(status.Get_tag(), ctx.tag)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        greq.Wait()

    def testAll2(self):
        greq = MPI.Grequest.Start(None, None, None)
        self.assertFalse(greq.Test())
        greq.Cancel()
        greq.Complete()
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(status.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        greq.Wait()

    def testPyCompleteTest(self):
        greq = MPI.Grequest.Start()
        self.assertFalse(greq.Test())
        greq.cancel()
        greq.complete(42)
        status = MPI.Status()
        flag, result = greq.test(status)
        self.assertTrue(flag)
        self.assertEqual(result, 42)
        self.assertEqual(status.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(status.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        obj = greq.wait()
        self.assertIsNone(obj)

    def testPyCompleteWait(self):
        greq = MPI.Grequest.Start()
        self.assertFalse(greq.Test())
        greq.cancel()
        greq.complete(42)
        status = MPI.Status()
        result = greq.wait(status)
        self.assertEqual(result, 42)
        self.assertEqual(status.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(status.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        flag, obj = greq.test()
        self.assertTrue(flag)
        self.assertIsNone(obj)


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    """
    #define MPICH_SKIP_MPICXX 1
    #define OMPI_SKIP_MPICXX  1
    """

cdef extern from "<mpi.h>" nogil:
    """
    #include "lib-mpi.h"
    #include "pyapicompat.h"
    #include "pympivendor.h"
    #include "pympicommctx.h"
    """

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    enum: PY_VERSION_HEX
    ctypedef ssize_t Py_intptr_t
    ctypedef size_t  Py_uintptr_t

cdef extern from "Python.h":
    const char *Py_GETENV(const char[]) nogil

cdef extern from "Python.h":
    void *PyExc_RuntimeError
    void *PyExc_NotImplementedError
    void PyErr_SetNone(object)
    void PyErr_SetObject(object, object)
    void PyErr_DisplayException(object)
    int  PyErr_WarnFormat(object, Py_ssize_t, const char[], ...) except -1
    void PySys_WriteStderr(const char[], ...)

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    """
    #if defined(PYPY_VERSION)
    #  define PyMPI_RUNTIME_PYPY 1
    #else
    #  define PyMPI_RUNTIME_PYPY 0
    #endif
    """
    enum: PYPY "PyMPI_RUNTIME_PYPY"

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    """
    #if !defined(PyMPI_USE_MATCHED_RECV)
    #  if defined(PyMPI_HAVE_MPI_Mprobe) && defined(PyMPI_HAVE_MPI_Mrecv)
    #    if defined(MPI_VERSION) && MPI_VERSION >= 3
    #      define PyMPI_USE_MATCHED_RECV 1
    #    endif
    #  endif
    #endif
    #if !defined(PyMPI_USE_MATCHED_RECV)
    #  define PyMPI_USE_MATCHED_RECV 0
    #endif
    """
    enum: USE_MATCHED_RECV "PyMPI_USE_MATCHED_RECV"

ctypedef struct Options:
    bint      initialize
    bint      threads
    int       thread_level
    bint      finalize
    bint      fast_reduce
    bint      recv_mprobe
    MPI_Count irecv_bufsz
    int       errors

cdef Options options
options.initialize = 1
options.threads = 1
options.thread_level = MPI_THREAD_MULTIPLE
options.finalize = 1
options.fast_reduce = 1
options.recv_mprobe = 1
options.irecv_bufsz = 32768
options.errors = 1

cdef object getOpt(object rc, const char name[], object value):
    cdef bytes bname = b"MPI4PY_RC_" + name.upper()
    cdef const char *cname  = <const char*>bname
    cdef const char *cvalue = Py_GETENV(cname)
    if cvalue == NULL:
        return getattr(rc, pystr(name), value)
    if cstr_is_uint(cvalue) and (type(value) is int):
        value = int(pystr(cvalue))
    elif cstr_is_bool(cvalue):
        value = <bint>cstr2bool(cvalue)
    else:
        value = pystr(cvalue).lower()
    try:
        setattr(rc, pystr(name), value)
    except:  # noqa
        pass
    return value

cdef int warnOpt(const char name[], object value) except -1:
    value = PyUnicode_AsUTF8String(repr(value))
    PyErr_WarnFormat(
        RuntimeWarning, 1,
        b"mpi4py.rc.%s: unexpected value %.200s",
        name, <const char*>value,
    )
    return 0

cdef int getOptions(Options* opts) except -1:
    opts.initialize = 1
    opts.threads = 1
    opts.thread_level = MPI_THREAD_MULTIPLE
    opts.finalize = 1
    opts.fast_reduce = 1
    opts.recv_mprobe = USE_MATCHED_RECV
    opts.irecv_bufsz = 32768
    opts.errors = 1
    #
    cdef object rc
    try:
        from . import rc
    except (ImportError, ImportWarning):
        rc = None
    #
    cdef object initialize   = getOpt(rc, b"initialize"   , True        )
    cdef object threads      = getOpt(rc, b"threads"      , True        )
    cdef object thread_level = getOpt(rc, b"thread_level" , 'multiple'  )
    cdef object finalize     = getOpt(rc, b"finalize"     , None        )
    cdef object fast_reduce  = getOpt(rc, b"fast_reduce"  , True        )
    cdef object recv_mprobe  = getOpt(rc, b"recv_mprobe"  , True        )
    cdef object irecv_bufsz  = getOpt(rc, b"irecv_bufsz"  , 32768       )
    cdef object errors       = getOpt(rc, b"errors"       , 'exception' )
    #
    if initialize in (True, 'yes'):
        opts.initialize = 1
    elif initialize in (False, 'no'):
        opts.initialize = 0
    else:
        warnOpt(b"initialize", initialize)
    #
    if threads in (True, 'yes'):
        opts.threads = 1
    elif threads in (False, 'no'):
        opts.threads = 0
    else:
        warnOpt(b"threads", threads)
    #
    if thread_level == 'single':
        opts.thread_level = MPI_THREAD_SINGLE
    elif thread_level == 'funneled':
        opts.thread_level = MPI_THREAD_FUNNELED
    elif thread_level == 'serialized':
        opts.thread_level = MPI_THREAD_SERIALIZED
    elif thread_level == 'multiple':
        opts.thread_level = MPI_THREAD_MULTIPLE
    else:
        warnOpt(b"thread_level", thread_level)
    #
    if finalize is None:
        opts.finalize = opts.initialize
    elif finalize in (True, 'yes'):
        opts.finalize = 1
    elif finalize in (False, 'no'):
        opts.finalize = 0
    else:
        warnOpt(b"finalize", finalize)
    #
    if fast_reduce in (True, 'yes'):
        opts.fast_reduce = 1
    elif fast_reduce in (False, 'no'):
        opts.fast_reduce = 0
    else:
        warnOpt(b"fast_reduce", fast_reduce)
    #
    if recv_mprobe in (True, 'yes'):
        opts.recv_mprobe = 1 and USE_MATCHED_RECV
    elif recv_mprobe in (False, 'no'):
        opts.recv_mprobe = 0
    else:
        warnOpt(b"recv_mprobe", recv_mprobe)
    #
    if type(irecv_bufsz) is int and irecv_bufsz >= 0:
        opts.irecv_bufsz = irecv_bufsz
    else:
        warnOpt(b"irecv_bufsz", irecv_bufsz)
    #
    if errors == 'default':
        opts.errors = 0
    elif errors == 'exception':
        opts.errors = 1
    elif errors == 'abort':
        opts.errors = 2 if MPI_ERRORS_ABORT != MPI_ERRHANDLER_NULL else 3
    elif errors == 'fatal':
        opts.errors = 3
    else:
        warnOpt(b"errors", errors)
    #
    return 0

# -----------------------------------------------------------------------------

cdef int warn_environ(const char envvar[]) except -1 with gil:
    PyErr_WarnFormat(
        RuntimeWarning, 1,
        b"environment variable %s: "
        b"unexpected value '%.200s'",
        envvar, getenv(envvar),
    )

cdef int warn_mpiexec(
    const char envvar[],
    const char vendor[],
) except -1 with gil:
    PyErr_WarnFormat(
        RuntimeWarning, 1,
        b"suspicious MPI execution environment\n"
        b"Your environment has %s=%.200s set, "
        b"but mpi4py was built with %s.\n"
        b"You may be using `mpiexec` or `mpirun` "
        b"from a different MPI implementation.",
        envvar, getenv(envvar), vendor,
    )

cdef int check_mpiexec() except -1 nogil:
    cdef int ierr, size = 0
    ierr = MPI_Comm_size(MPI_COMM_WORLD, &size)
    if ierr != MPI_SUCCESS: return 0
    if size > 1: return 0

    cdef int check = 1
    cdef const char *ename  = b"MPI4PY_CHECK_MPIEXEC"
    cdef const char *value  = Py_GETENV(ename)
    if value != NULL: check = cstr2bool(value)
    if check == -1: warn_environ(ename)
    if check <=  0: return 0

    cdef const char *vendor = NULL
    <void>PyMPI_Get_vendor(&vendor, NULL, NULL, NULL)

    cdef bint MPICH   = (strncmp(b"MPICH", vendor, 6) == 0)
    cdef bint IMPI    = (strncmp(b"Intel MPI", vendor, 10) == 0)
    cdef bint OPENMPI = (strncmp(b"Open MPI", vendor, 9) == 0)
    cdef const char *mpich   = b"HYDI_CONTROL_FD"
    cdef const char *impi    = b"I_MPI_HYDRA_TOPOLIB"
    cdef const char *pmi_sz  = b"PMI_SIZE"
    cdef const char *openmpi = b"OMPI_COMM_WORLD_SIZE"
    cdef const char *bad_env = NULL
    MPICH |= (strncmp(b"MVAPICH", vendor, 8) == 0)
    MPICH |= (strncmp(b"MVAPICH2", vendor, 9) == 0)
    if MPICH:
        if getenv(mpich) == NULL and getenv(pmi_sz) == NULL:      # ~> mpich
            if getenv(openmpi) != NULL:                           # ~> mpich
                bad_env = openmpi                                 # ~> mpich
    if IMPI:
        if getenv(impi) == NULL and getenv(pmi_sz) == NULL:       # ~> impi
            if getenv(openmpi) != NULL:                           # ~> impi
                bad_env = openmpi                                 # ~> impi
    if OPENMPI:
        if getenv(openmpi) == NULL:                               # ~> openmpi
            if getenv(mpich) != NULL and getenv(pmi_sz) != NULL:  # ~> openmpi
                bad_env = pmi_sz                                  # ~> openmpi
            if getenv(impi ) != NULL and getenv(pmi_sz) != NULL:  # ~> openmpi
                bad_env = pmi_sz                                  # ~> openmpi
    if bad_env != NULL:
        warn_mpiexec(bad_env, vendor)
    return 0

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int Py_AtExit(void (*)() noexcept nogil)

cdef int bootstrap() except -1:
    # Get options from 'mpi4py.rc' module
    getOptions(&options)
    # Cleanup at (the very end of) Python exit
    if Py_AtExit(atexit) < 0:
        PySys_WriteStderr(                                   # ~> uncovered
            b"WARNING: %s\n",                                # ~> uncovered
            b"could not register cleanup with Py_AtExit()",  # ~> uncovered
        )
    # Do we have to initialize MPI?
    cdef int initialized = 1
    <void>MPI_Initialized(&initialized)
    if initialized:
        options.finalize = 0  # ~> TODO
        return 0              # ~> TODO
    if not options.initialize:
        return 0
    # MPI initialization
    cdef int ierr = MPI_SUCCESS
    cdef int required = MPI_THREAD_SINGLE
    cdef int provided = MPI_THREAD_SINGLE
    if options.threads:
        required = options.thread_level
        ierr = MPI_Init_thread(NULL, NULL, required, &provided)
        if ierr != MPI_SUCCESS:
            raise RuntimeError(               # ~> uncovered
                f"MPI_Init_thread() failed "  # ~> uncovered
                f"[error code: {ierr}]")      # ~> uncovered
    else:
        ierr = MPI_Init(NULL, NULL)
        if ierr != MPI_SUCCESS:
            raise RuntimeError(               # ~> uncovered
                f"MPI_Init() failed "         # ~> uncovered
                f"[error code: {ierr}]")      # ~> uncovered
    return 0


@cython.linetrace(False)
cdef inline int mpi_active() noexcept nogil:
    cdef int ierr = MPI_SUCCESS
    # MPI initialized ?
    cdef int initialized = 0
    ierr = MPI_Initialized(&initialized)
    if not initialized or ierr != MPI_SUCCESS: return 0
    # MPI finalized ?
    cdef int finalized = 1
    ierr = MPI_Finalized(&finalized)
    if finalized or ierr != MPI_SUCCESS: return 0
    # MPI should be active ...
    return 1


cdef int mpi_version    = 0
cdef int mpi_subversion = 0
cdef int mpi_numversion = 0

cdef int initialize() except -1 nogil:
    global mpi_version, mpi_subversion, mpi_numversion
    <void> MPI_Get_version(&mpi_version, &mpi_subversion)
    mpi_numversion = 10 * mpi_version + mpi_subversion
    if not mpi_active(): return 0
    check_mpiexec()
    options_set_errhandler(MPI_COMM_SELF)
    options_set_errhandler(MPI_COMM_WORLD)
    return 0


@cython.linetrace(False)
cdef void finalize() noexcept nogil:
    if not mpi_active(): return
    <void>PyMPI_Commctx_finalize()


cdef int abort_status = 0


@cython.linetrace(False)
cdef void atexit() noexcept nogil:
    if not mpi_active(): return
    if abort_status:
        <void>MPI_Abort(MPI_COMM_WORLD, abort_status)
    finalize()
    if options.finalize:
        <void>MPI_Finalize()


def _set_abort_status(int status: int) -> None:
    """
    Helper for ``python -m mpi4py.run ...``.
    """
    global abort_status
    abort_status = status


# -----------------------------------------------------------------------------

# Raise exceptions without adding to traceback

cdef extern from * nogil:
    enum: PyMPI_ERR_UNAVAILABLE

cdef object MPIException = <object>PyExc_RuntimeError


@cython.linetrace(False)
cdef int PyMPI_Raise(int ierr) except -1 with gil:
    if ierr == PyMPI_ERR_UNAVAILABLE:
        PyErr_SetObject(<object>PyExc_NotImplementedError, None)
        return 0
    if (<void*>MPIException) == NULL:
        PyErr_SetObject(<object>PyExc_RuntimeError, <long>ierr)
        return 0
    PyErr_SetObject(MPIException, <long>ierr)
    return 0


@cython.linetrace(False)
cdef inline int CHKERR(int ierr) except -1 nogil:
    if ierr == MPI_SUCCESS: return 0
    PyMPI_Raise(ierr)
    return -1


cdef int PyMPI_HandleException(object exc) noexcept:
    PyErr_DisplayException(exc)
    if (<void*>MPIException) != NULL:
        if isinstance(exc, Exception):
            return <int>exc
    return MPI_ERR_OTHER

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    # PyPy: Py_IsInitialized() cannot be called without the GIL
    int _Py_IsInitialized"Py_IsInitialized"() noexcept nogil

cdef object _py_module_sentinel = None


@cython.linetrace(False)
cdef inline int py_module_alive() noexcept nogil:
    return NULL != <void *>_py_module_sentinel


@cython.linetrace(False)
cdef inline int Py_IsInitialized() noexcept nogil:
    if PYPY and not py_module_alive(): return 0
    return _Py_IsInitialized()

# -----------------------------------------------------------------------------



import mpiunittest as unittest

from mpi4py import MPI

try:
    import array
except ImportError:
    array = None


def asarray(typecode, data):
    def tobytes(s):
        return memoryview(s).tobytes()

    frombytes = array.array.frombytes
    a = array.array(typecode, [])
    frombytes(a, tobytes(data))
    return a


def mysum_obj(a, b):
    for i in range(len(a)):
        b[i] = a[i] + b[i]
    return b


def mysum_buf(a, b, dt):
    assert dt == MPI.INT
    assert len(a) == len(b)
    b[:] = mysum_obj(asarray("i", a), asarray("i", b))


def mysum(ba, bb, dt):
    if dt is None:
        return mysum_obj(ba, bb)
    else:
        return mysum_buf(ba, bb, dt)


def mybor(a, b, dt):
    assert dt == MPI.BYTE
    assert len(a) == len(b)
    for i in range(len(a)):
        b[i] = a[i] | b[i]


class TestOp(unittest.TestCase):
    #
    def testConstructor(self):
        op = MPI.Op()
        self.assertFalse(op)
        self.assertEqual(op, MPI.OP_NULL)

    @unittest.skipIf(array is None, "array")
    def testCreate(self):
        for comm in [MPI.COMM_SELF, MPI.COMM_WORLD]:
            for commute in [True, False]:
                for N in range(4):
                    myop = MPI.Op.Create(mysum, commute)
                    self.assertFalse(myop.is_predefined)
                    try:
                        size = comm.Get_size()
                        rank = comm.Get_rank()
                        a = array.array(
                            "i", [i * (rank + 1) for i in range(N)]
                        )
                        b = array.array("i", [0] * len(a))
                        comm.Allreduce([a, MPI.INT], [b, MPI.INT], myop)
                        scale = sum(range(1, size + 1))
                        for i in range(N):
                            self.assertEqual(b[i], scale * i)
                        res = myop(a, b)
                        self.assertIs(res, b)
                        for i in range(N):
                            self.assertEqual(b[i], a[i] + scale * i)
                        myop2 = MPI.Op(myop)
                        a = array.array("i", [1] * N)
                        b = array.array("i", [2] * N)
                        res = myop2(a, b)
                        self.assertIs(res, b)
                        for i in range(N):
                            self.assertEqual(b[i], 3)
                        myop2 = None
                    finally:
                        myop.Free()

    def testCreateMany(self):
        MAX_USER_OP = 32  # max user-defined operations
        # create
        ops = []
        for _i in range(MAX_USER_OP):
            o = MPI.Op.Create(mysum)
            ops.append(o)
        with self.assertRaises(RuntimeError):
            MPI.Op.Create(mysum)
        # cleanup
        for o in ops:
            o.Free()
        # another round
        ops = []
        for _i in range(MAX_USER_OP):
            op = MPI.Op.Create(mybor)
            ops.append(op)
        with self.assertRaises(RuntimeError):
            MPI.Op.Create(mybor)
        # local reductions
        try:
            b1 = bytearray([2] * 3)
            b2 = bytearray([4] * 3)
            for op in ops:
                ibuf = [b1, MPI.BYTE]
                obuf = [b2, MPI.BYTE]
                op.Reduce_local(ibuf, obuf)
                for c1, c2 in zip(b1, b2):
                    self.assertEqual(c1, 2)
                    self.assertEqual(c2, 6)
        except NotImplementedError:
            pass
        # pickling support
        for op in ops:
            op.__reduce__()
            if op.toint() != -1:
                clon = MPI.Op.fromint(op.toint())
                with self.assertRaises(ValueError):
                    clon.__reduce__()
        # cleanup
        for op in ops:
            op.Free()

    def _test_call(self, op, args, res):
        self.assertEqual(op(*args), res)
        self.assertEqual(MPI.Op(op)(*args), res)

    def testCall(self):
        self.assertRaises(TypeError, MPI.OP_NULL)
        self.assertRaises(TypeError, MPI.OP_NULL, None)
        self.assertRaises(ValueError, MPI.OP_NULL, None, None)
        self._test_call(MPI.MIN, (2, 3), 2)
        self._test_call(MPI.MAX, (2, 3), 3)
        self._test_call(MPI.SUM, (2, 3), 5)
        self._test_call(MPI.PROD, (2, 3), 6)
        for x in (False, True):
            for y in (False, True):
                self._test_call(MPI.LAND, (x, y), x and y)
                self._test_call(MPI.LOR, (x, y), x or y)
                self._test_call(MPI.LXOR, (x, y), x ^ y)
        for x in range(5):
            for y in range(5):
                self._test_call(MPI.BAND, (x, y), x & y)
                self._test_call(MPI.BOR, (x, y), x | y)
                self._test_call(MPI.BXOR, (x, y), x ^ y)
        if MPI.REPLACE:
            self._test_call(MPI.REPLACE, (2, 3), 3)
            self._test_call(MPI.REPLACE, (3, 2), 2)
        if MPI.NO_OP:
            self._test_call(MPI.NO_OP, (2, 3), 2)
            self._test_call(MPI.NO_OP, (3, 2), 3)

    def testMinMax(self):
        x = [1]
        y = [1]
        res = MPI.MIN(x, y)
        self.assertEqual(res, x)
        res = MPI.MAX(x, y)
        self.assertEqual(res, x)

    def testMinMaxLoc(self):
        x = [1]
        i = [2]
        u = [x, i]
        y = [2]
        j = [1]
        v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        res = MPI.MINLOC(v, u)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        res = MPI.MAXLOC(u, v)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)
        res = MPI.MAXLOC(v, u)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)
        #
        x = [1]
        i = 0
        u = [x, i]
        y = [1]
        j = 1
        v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        res = MPI.MAXLOC(u, v)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        #
        x = [1]
        i = 1
        u = [x, i]
        y = [1]
        j = 0
        v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)
        res = MPI.MAXLOC(u, v)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)
        #
        x = [1]
        i = [0]
        u = [x, i]
        y = [1]
        j = [1]
        v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        res = MPI.MAXLOC(u, v)
        self.assertIs(res[0], x)
        self.assertIs(res[1], i)
        #
        x = [1]
        i = [1]
        u = [x, i]
        y = [1]
        j = [0]
        v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)
        res = MPI.MAXLOC(u, v)
        self.assertIs(res[0], y)
        self.assertIs(res[1], j)

    @unittest.skipMPI("openmpi(<=1.8.1)")
    def testIsCommutative(self):
        try:
            MPI.SUM.Is_commutative()
        except NotImplementedError:
            self.skipTest("mpi-op-is_commutative")
        ops = [
            MPI.MAX,
            MPI.MIN,
            MPI.SUM,
            MPI.PROD,
            MPI.LAND,
            MPI.BAND,
            MPI.LOR,
            MPI.BOR,
            MPI.LXOR,
            MPI.BXOR,
            MPI.MAXLOC,
            MPI.MINLOC,
        ]
        for op in ops:
            flag = op.Is_commutative()
            self.assertEqual(flag, op.is_commutative)
            self.assertTrue(flag)

    @unittest.skipMPI("openmpi(<=1.8.1)")
    @unittest.skipMPI("mpich(==3.4.1)")
    def testIsCommutativeExtra(self):
        try:
            MPI.SUM.Is_commutative()
        except NotImplementedError:
            self.skipTest("mpi-op-is_commutative")
        ops = [MPI.REPLACE, MPI.NO_OP]
        for op in ops:
            if not op:
                continue
            flag = op.Is_commutative()
            self.assertEqual(flag, op.is_commutative)
            # self.assertFalse(flag)

    def testIsPredefined(self):
        self.assertTrue(MPI.OP_NULL.is_predefined)
        ops = [
            MPI.MAX,
            MPI.MIN,
            MPI.SUM,
            MPI.PROD,
            MPI.LAND,
            MPI.BAND,
            MPI.LOR,
            MPI.BOR,
            MPI.LXOR,
            MPI.BXOR,
            MPI.MAXLOC,
            MPI.MINLOC,
        ]
        for op in ops:
            self.assertTrue(op.is_predefined)

    def testPicklePredefined(self):
        from pickle import dumps, loads

        ops = [
            MPI.MAX,
            MPI.MIN,
            MPI.SUM,
            MPI.PROD,
            MPI.LAND,
            MPI.BAND,
            MPI.LOR,
            MPI.BOR,
            MPI.LXOR,
            MPI.BXOR,
            MPI.MAXLOC,
            MPI.MINLOC,
            MPI.OP_NULL,
            MPI.NO_OP,
        ]
        for op in ops:
            newop = loads(dumps(op))
            self.assertIs(newop, op)
            newop = loads(dumps(MPI.Op(op)))
            self.assertIsNot(newop, op)
            self.assertEqual(newop, op)

    def testPickleUserDefined(self):
        from pickle import dumps, loads

        for commute in [True, False]:
            myop1 = MPI.Op.Create(mysum, commute)
            myop2 = loads(dumps(myop1))
            self.assertNotEqual(myop1, myop2)
            myop1.Free()
            self.assertEqual(myop2([2], [3]), [5])
            self.assertEqual(myop2.Is_commutative(), commute)
            myop2.Free()


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

ctypedef fused PyMPI_attach_buffer_type:
    Py_uintptr_t
    Comm
    Session

cdef dict _mpi_buffer_comm = {}
cdef dict _mpi_buffer_session = {}


cdef inline object attach_buffer(
    object buf,
    void **p,
    MPI_Count *n,
):
    cdef void *bptr = MPI_BUFFER_AUTOMATIC
    cdef MPI_Aint blen = 0
    if buf is None or is_BUFFER_AUTOMATIC(buf):
        buf = __BUFFER_AUTOMATIC__
    else:
        buf = asbuffer_w(buf, &bptr, &blen)
    p[0] = bptr
    n[0] = blen
    return buf

cdef inline int detach_buffer_set(
    PyMPI_attach_buffer_type obj,
    object buf,
) except -1:
    cdef Py_uintptr_t handle
    if PyMPI_attach_buffer_type is Py_uintptr_t:
        handle = <Py_uintptr_t>obj
        <void> handle  # unused
        _mpi_buffer_comm[None] = buf
    if PyMPI_attach_buffer_type is Comm:
        handle = <Py_uintptr_t>obj.ob_mpi  # ~> MPI-4.1
        _mpi_buffer_comm[handle] = buf     # ~> MPI-4.1
    if PyMPI_attach_buffer_type is Session:
        handle = <Py_uintptr_t>obj.ob_mpi  # ~> MPI-4.1
        _mpi_buffer_session[handle] = buf  # ~> MPI-4.1
    return 0

cdef inline object detach_buffer_get(
    PyMPI_attach_buffer_type obj,
    void *p,
    MPI_Count n,
):
    cdef Py_uintptr_t handle
    cdef buffer buf = <buffer>None
    if PyMPI_attach_buffer_type is Py_uintptr_t:
        handle = <Py_uintptr_t>obj
        <void> handle  # unused
        buf = <buffer>_mpi_buffer_comm.pop(None, None)
    if PyMPI_attach_buffer_type is Comm:
        handle = <Py_uintptr_t>obj.ob_mpi                 # ~> MPI-4.1
        buf = <buffer>_mpi_buffer_comm.pop(handle, None)  # ~> MPI-4.1
    if PyMPI_attach_buffer_type is Session:
        handle = <Py_uintptr_t>obj.ob_mpi                    # ~> MPI-4.1
        buf = <buffer>_mpi_buffer_session.pop(handle, None)  # ~> MPI-4.1
    if p == MPI_BUFFER_AUTOMATIC:
        return __BUFFER_AUTOMATIC__  # ~> MPI-4.1
    if buf is not None and buf.view.buf == p and buf.view.obj != NULL:
        return <object> buf.view.obj
    return mpibuf(p, n)

# -----------------------------------------------------------------------------



import os
import pathlib
import sys
import tempfile

import mpiunittest as unittest

from mpi4py import MPI


class TestErrhandler(unittest.TestCase):
    #
    def testPredefined(self):
        self.assertFalse(MPI.ERRHANDLER_NULL)
        self.assertTrue(MPI.ERRORS_ARE_FATAL)
        self.assertTrue(MPI.ERRORS_RETURN)
        if MPI.Get_version() >= (4, 0):
            self.assertTrue(MPI.ERRORS_ABORT)
        elif MPI.ERRORS_ABORT != MPI.ERRHANDLER_NULL:
            self.assertTrue(MPI.ERRORS_ABORT)
        else:
            self.assertFalse(MPI.ERRORS_ABORT)

    def testPickle(self):
        from pickle import dumps, loads

        for errhandler in [
            MPI.ERRHANDLER_NULL,
            MPI.ERRORS_ARE_FATAL,
            MPI.ERRORS_RETURN,
            MPI.ERRORS_ABORT,
        ]:
            if not errhandler:
                continue
            errh = loads(dumps(errhandler))
            self.assertIs(errh, errhandler)
            errh = loads(dumps(MPI.Errhandler(errhandler)))
            self.assertIsNot(errh, errhandler)
            self.assertEqual(errh, errhandler)


class BaseTestErrhandler:
    #
    def testCreate(self):
        MAX_USER_EH = 32  # max user-defined error handlers
        mpiobj = self.mpiobj
        index = None
        called = check = False

        def get_errhandler_fn(idx):
            def errhandler_fn(arg, err):
                nonlocal called, check
                called = check = True
                check &= arg == mpiobj
                check &= err == MPI.ERR_OTHER
                check &= idx == index

            return errhandler_fn

        def check_handle(eh):
            cint = eh.toint()
            if cint != -1:
                clon = type(eh).fromint(cint)
                self.assertEqual(eh, clon)
            fint = eh.py2f()
            if fint != -1:
                clon = type(eh).f2py(fint)
                self.assertEqual(eh, clon)

        errhandlers = []
        for index in range(MAX_USER_EH):
            try:
                fn = get_errhandler_fn(index)
                eh = type(mpiobj).Create_errhandler(fn)
                errhandlers.append(eh)
            except NotImplementedError:
                clsname = type(mpiobj).__name__.lower()
                self.skipTest(f"mpi-{clsname}-create_errhandler")
        with self.assertRaises(RuntimeError):
            type(mpiobj).Create_errhandler(lambda _arg, _err: None)

        for eh in errhandlers:
            self.assertTrue(eh)
            check_handle(eh)
            with self.assertRaises(ValueError):
                eh.__reduce__()

        eh_orig = mpiobj.Get_errhandler()
        try:
            for i, eh in enumerate(errhandlers):
                index = i
                called = check = False
                mpiobj.Set_errhandler(eh)
                try:
                    mpiobj.Call_errhandler(MPI.SUCCESS)
                    mpiobj.Call_errhandler(MPI.ERR_OTHER)
                except NotImplementedError:
                    if MPI.Get_version() >= (2, 0):
                        raise
                else:
                    self.assertTrue(called)
                    self.assertTrue(check)
                finally:
                    mpiobj.Set_errhandler(eh_orig)
        finally:
            for eh in errhandlers:
                eh.Free()

    def testCall(self):
        mpiobj = self.mpiobj
        mpiobj.Set_errhandler(MPI.ERRORS_RETURN)
        try:
            mpiobj.Call_errhandler(MPI.SUCCESS)
            mpiobj.Call_errhandler(MPI.ERR_OTHER)
        except NotImplementedError:
            if MPI.Get_version() >= (2, 0):
                raise
            clsname = type(mpiobj).__name__.lower()
            self.skipTest(f"mpi-{clsname}-call_errhandler")

    def testGetFree(self):
        mpiobj = self.mpiobj
        errhdls = []
        for _i in range(100):
            e = mpiobj.Get_errhandler()
            errhdls.append(e)
        for e in errhdls:
            e.Free()
        for e in errhdls:
            self.assertEqual(e, MPI.ERRHANDLER_NULL)

    def _run_test_get_set(self, errhandler):
        mpiobj = self.mpiobj
        errhdl_1 = mpiobj.Get_errhandler()
        self.assertNotEqual(errhdl_1, MPI.ERRHANDLER_NULL)
        mpiobj.Set_errhandler(errhandler)
        errhdl_2 = mpiobj.Get_errhandler()
        self.assertEqual(errhdl_2, errhandler)
        errhdl_2.Free()
        self.assertEqual(errhdl_2, MPI.ERRHANDLER_NULL)
        mpiobj.Set_errhandler(errhdl_1)
        errhdl_1.Free()
        self.assertEqual(errhdl_1, MPI.ERRHANDLER_NULL)

    def testErrorsReturn(self):
        self._run_test_get_set(MPI.ERRORS_RETURN)

    def testErrorsFatal(self):
        self._run_test_get_set(MPI.ERRORS_ARE_FATAL)

    @unittest.skipUnless(MPI.ERRORS_ABORT, "mpi-errors-abort")
    @unittest.skipMPI("mpich(<4.1.0)")
    @unittest.skipMPI("openmpi(<5.0.0)")
    @unittest.skipMPI("impi(<2021.14.0)")
    def testErrorsAbort(self):
        self._run_test_get_set(MPI.ERRORS_ABORT)


class TestErrhandlerComm(BaseTestErrhandler, unittest.TestCase):
    #
    def setUp(self):
        self.mpiobj = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.mpiobj.Free()


@unittest.skipMPI("openmpi(<4.1.0)", sys.platform == "darwin")
class TestErrhandlerWin(BaseTestErrhandler, unittest.TestCase):
    #
    def setUp(self):
        try:
            self.mpiobj = MPI.Win.Create(
                MPI.BOTTOM, 1, MPI.INFO_NULL, MPI.COMM_SELF
            )
        except NotImplementedError:
            self.skipTest("mpi-win")
        except MPI.Exception:
            self.skipTest("mpi-win")

    def tearDown(self):
        self.mpiobj.Free()

    def testCall(self):
        super().testCall()


class TestErrhandlerFile(BaseTestErrhandler, unittest.TestCase):
    #
    def setUp(self):
        rank = MPI.COMM_WORLD.Get_rank()
        fd, filename = tempfile.mkstemp(prefix="mpi4py-", suffix=f"-{rank}")
        os.close(fd)
        amode = MPI.MODE_WRONLY | MPI.MODE_CREATE | MPI.MODE_DELETE_ON_CLOSE
        try:
            self.mpiobj = MPI.File.Open(
                MPI.COMM_SELF, filename, amode, MPI.INFO_NULL
            )
        except NotImplementedError:
            try:
                pathlib.Path(filename).unlink()
            except OSError:
                pass
            self.skipTest("mpi-file")

    def tearDown(self):
        self.mpiobj.Close()

    @unittest.skipMPI("msmpi")
    def testCall(self):
        super().testCall()


class TestErrhandlerSession(BaseTestErrhandler, unittest.TestCase):
    #
    def setUp(self):
        try:
            self.mpiobj = MPI.Session.Init()
        except NotImplementedError:
            self.skipTest("mpi-session")

    def tearDown(self):
        self.mpiobj.Finalize()


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------

import sys

from mpi4py import MPI


class Counter:
    def __init__(self, comm):
        assert not comm.Is_inter()
        self.comm = comm.Dup()
        # start counter process
        script = __file__
        self.child = self.comm.Spawn(sys.executable, [script, "--child"], 1)

    def free(self):
        self.comm.Barrier()
        # stop counter process
        rank = self.child.Get_rank()
        if rank == 0:
            self.child.send(None, 0, 1)
        self.child.Disconnect()
        #
        self.comm.Free()

    def next(self):
        #
        incr = 1
        self.child.send(incr, 0, 0)
        ival = self.child.recv(None, 0, 0)
        nxtval = ival
        #
        return nxtval


# --------------------------------------------------------------------


def _counter_child():
    parent = MPI.Comm.Get_parent()
    assert parent != MPI.COMM_NULL
    try:
        counter = 0
        status = MPI.Status()
        any_src, any_tag = MPI.ANY_SOURCE, MPI.ANY_TAG
        while True:  # server loop
            incr = parent.recv(None, any_src, any_tag, status)
            if status.tag == 1:
                break
            parent.send(counter, status.source, 0)
            counter += incr
    finally:
        parent.Disconnect()


if __name__ == "__main__":
    if (
        len(sys.argv) > 1
        and sys.argv[0] == __file__
        and sys.argv[1] == "--child"
    ):
        _counter_child()
        sys.exit(0)

# --------------------------------------------------------------------


def test():
    vals = []
    counter = Counter(MPI.COMM_WORLD)
    for _i in range(5):
        c = counter.next()
        vals.append(c)
    counter.free()
    #
    vals = MPI.COMM_WORLD.allreduce(vals)
    assert sorted(vals) == list(range(len(vals)))


if __name__ == "__main__":
    test()

# --------------------------------------------------------------------



import logging
import os
import platform
import shlex
import shutil
import sys
from collections import OrderedDict
from configparser import ConfigParser, Error as ConfigParserError
from types import SimpleNamespace

# ruff: noqa: PTH112, PTH113, PTH117, PTH118, PTH119
# ruff: noqa: PTH120, PTH123

_logger = logging.getLogger("mpiconfig")
_logger.setLevel(logging.INFO)

# ruff: noqa: G010
_log = SimpleNamespace()
_log.log = _logger.log
_log.debug = _logger.debug
_log.info = _logger.info
_log.warn = _logger.warning
_log.error = _logger.error
_log.fatal = _logger.fatal


class Config:
    def __init__(self, log=None):
        self.log = log or _log
        self.section = None
        self.filename = None
        self.compiler_info = OrderedDict((
            ("mpicc", None),
            ("mpicxx", None),
            ("mpild", None),
        ))
        self.library_info = OrderedDict((
            ("define_macros", []),
            ("undef_macros", []),
            ("include_dirs", []),
            ("libraries", []),
            ("library_dirs", []),
            ("runtime_library_dirs", []),
            ("extra_compile_args", []),
            ("extra_link_args", []),
            ("extra_objects", []),
        ))

    def __bool__(self):
        for v in self.compiler_info.values():
            if v:
                return True
        for v in self.library_info.values():
            if v:
                return True
        return False

    def get(self, k, d=None):
        if k in self.compiler_info:
            return self.compiler_info[k]
        if k in self.library_info:
            return self.library_info[k]
        return d

    def info(self, log=None):
        if log is None:
            log = self.log
        mpicc = self.compiler_info.get("mpicc")
        mpicxx = self.compiler_info.get("mpicxx")
        mpild = self.compiler_info.get("mpild")
        if mpicc:
            log.info("MPI C compiler:    %s", mpicc)
        if mpicxx:
            log.info("MPI C++ compiler:  %s", mpicxx)
        if mpild:
            log.info("MPI linker:        %s", mpild)

    def update(self, config, **more):
        if hasattr(config, "keys"):
            config = config.items()
        for option, value in config:
            if option in self.compiler_info:
                self.compiler_info[option] = value
            if option in self.library_info:
                self.library_info[option] = value
        if more:
            self.update(more)

    def setup(self, options, environ=None):
        if environ is None:
            environ = os.environ
        self.setup_library_info(options, environ)
        self.setup_compiler_info(options, environ)

    def setup_library_info(self, options, environ):
        filename = section = None
        mpiopt = getattr(options, "mpi", None)
        mpiopt = environ.get("MPICFG", mpiopt)
        mpiopt = environ.get("MPI4PY_BUILD_MPICFG", mpiopt)
        if mpiopt:
            if "," in mpiopt:
                section, filename = mpiopt.split(",", 1)
            elif ":" in mpiopt:
                filename, section = mpiopt.split(":", 1)
            elif os.path.isfile(mpiopt):
                filename = mpiopt
            else:
                section = mpiopt
        if not filename:
            filename = "mpi.cfg"
        if not section:
            section = "mpi"

        mach = platform.machine()
        arch = platform.architecture(None)[0]
        plat = sys.platform
        osnm = os.name
        if plat.startswith("linux"):
            plat = "linux"
        elif plat.startswith("sunos"):
            plat = "solaris"
        elif plat.startswith("win"):
            plat = "windows"
        suffixes = []
        suffixes.extend((
            plat + "-" + mach,
            plat + "-" + arch,
            plat,
            osnm + "-" + mach,
            osnm + "-" + arch,
            osnm,
            mach,
            arch,
        ))
        sections = [section + "-" + s for s in suffixes]
        sections += [section]
        self.load(filename, sections)
        if not self:
            self._setup_mpiabi()
        if not self:
            if os.name == "posix":
                self._setup_posix()
            if sys.platform == "win32":
                self._setup_windows()

    def _setup_mpiabi(self):
        MPI_ABI_STUBS = os.environ.get("MPI_ABI_STUBS")
        if MPI_ABI_STUBS:
            self.load("mpi.cfg", "mpiabi")
            self.filename = [MPI_ABI_STUBS]

    def _setup_posix(self):
        pass

    def _setup_windows(self):
        if self._setup_windows_impi():
            return
        if self._setup_windows_msmpi():
            return

    def _setup_windows_impi(self):
        I_MPI_ROOT = os.environ.get("I_MPI_ROOT")
        if not I_MPI_ROOT:
            return None
        IMPI_INC = os.path.join(I_MPI_ROOT, "include")
        IMPI_LIB = os.path.join(I_MPI_ROOT, "lib")
        kind = os.environ.get("I_MPI_LIBRARY_KIND") or "release"
        for subdirs in (("mpi", kind), (kind,)):
            if os.path.isdir(os.path.join(IMPI_LIB, *subdirs)):
                IMPI_LIB = os.path.join(IMPI_LIB, *subdirs)
                break
        ok = (
            os.path.isdir(I_MPI_ROOT)
            and os.path.isfile(os.path.join(IMPI_INC, "mpi.h"))
            and os.path.isfile(os.path.join(IMPI_LIB, "impi.lib"))
        )
        if not ok:
            return False
        IMPI_INC = os.path.normpath(IMPI_INC)
        IMPI_LIB = os.path.normpath(IMPI_LIB)
        self.library_info.update(
            include_dirs=[IMPI_INC],
            library_dirs=[IMPI_LIB],
            libraries=["impi"],
        )
        self.section = "impi"
        self.filename = [os.path.dirname(IMPI_INC)]
        return True

    def _setup_windows_msmpi(self):
        # Microsoft MPI (v7, v6, v5, v4)
        def msmpi_ver():
            try:
                try:
                    import winreg
                except ImportError:
                    import _winreg as winreg
                HKLM = winreg.HKEY_LOCAL_MACHINE
                subkey = r"SOFTWARE\Microsoft\MPI"
                with winreg.OpenKey(HKLM, subkey) as key:
                    for i in range(winreg.QueryInfoKey(key)[1]):
                        name, value, _type = winreg.EnumValue(key, i)
                        if name != "Version":
                            continue
                        major, minor = value.split(".")[:2]
                        return (int(major), int(minor))
            except Exception:  # noqa: S110
                pass
            MSMPI_VER = os.environ.get("MSMPI_VER")
            if MSMPI_VER:
                try:
                    major, minor = MSMPI_VER.split(".")[:2]
                    return (int(major), int(minor))
                except Exception:
                    raise RuntimeError(
                        f"invalid environment: MSMPI_VER={MSMPI_VER}"
                    ) from None
            return None

        def setup_msmpi(MSMPI_INC, MSMPI_LIB):
            from os.path import isfile, join

            ok = (
                MSMPI_INC
                and isfile(join(MSMPI_INC, "mpi.h"))
                and MSMPI_LIB
                and isfile(join(MSMPI_LIB, "msmpi.lib"))
            )
            if not ok:
                return False
            version = msmpi_ver()
            if version is not None:
                major, minor = version
                MSMPI_VER = hex((major << 8) | (minor & 0xFF))
                self.library_info.update(
                    define_macros=[("MSMPI_VER", MSMPI_VER)],
                )
            MSMPI_INC = os.path.normpath(MSMPI_INC)
            MSMPI_LIB = os.path.normpath(MSMPI_LIB)
            self.library_info.update(
                include_dirs=[MSMPI_INC],
                library_dirs=[MSMPI_LIB],
                libraries=["msmpi"],
            )
            self.section = "msmpi"
            self.filename = [os.path.dirname(MSMPI_INC)]
            return True

        arch = platform.architecture(None)[0][:2]
        # Look for Microsoft MPI in the environment
        MSMPI_INC = os.environ.get("MSMPI_INC")
        MSMPI_LIB = os.environ.get("MSMPI_LIB" + arch)
        MSMPI_LIB = MSMPI_LIB or os.environ.get("MSMPI_LIB")
        if setup_msmpi(MSMPI_INC, MSMPI_LIB):
            return True
        # Look for Microsoft MPI v7/v6/v5 in default install path
        for ProgramFiles in ("ProgramFiles", "ProgramFiles(x86)"):
            ProgramFiles = os.environ.get(ProgramFiles, "")
            archdir = {"32": "x86", "64": "x64"}[arch]
            MSMPI_DIR = os.path.join(ProgramFiles, "Microsoft SDKs", "MPI")
            MSMPI_INC = os.path.join(MSMPI_DIR, "Include")
            MSMPI_LIB = os.path.join(MSMPI_DIR, "Lib", archdir)
            if setup_msmpi(MSMPI_INC, MSMPI_LIB):
                return True
        # Look for Microsoft HPC Pack 2012 R2 in default install path
        for ProgramFiles in ("ProgramFiles", "ProgramFiles(x86)"):
            ProgramFiles = os.environ.get(ProgramFiles, "")
            archdir = {"32": "i386", "64": "amd64"}[arch]
            MSMPI_DIR = os.path.join(ProgramFiles, "Microsoft MPI")
            MSMPI_INC = os.path.join(MSMPI_DIR, "Inc")
            MSMPI_LIB = os.path.join(MSMPI_DIR, "Lib", archdir)
            if setup_msmpi(MSMPI_INC, MSMPI_LIB):
                return True
        # Microsoft MPI (legacy) and others
        ProgramFiles = os.environ.get("ProgramFiles", "")
        CCP_HOME = os.environ.get("CCP_HOME", "")
        for name, prefix, suffix in (
            ("msmpi", CCP_HOME, ""),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2012 R2"),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2012"),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2012 SDK"),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2008 R2"),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2008"),
            ("msmpi", ProgramFiles, "Microsoft HPC Pack 2008 SDK"),
        ):
            mpi_dir = os.path.join(prefix, suffix)
            if not mpi_dir or not os.path.isdir(mpi_dir):
                continue
            define_macros = []
            include_dir = os.path.join(mpi_dir, "include")
            library = "mpi"
            library_dir = os.path.join(mpi_dir, "lib")
            if name == "msmpi":
                include_dir = os.path.join(mpi_dir, "inc")
                library = "msmpi"
                arch = platform.architecture(None)[0]
                if arch == "32bit":
                    library_dir = os.path.join(library_dir, "i386")
                if arch == "64bit":
                    library_dir = os.path.join(library_dir, "amd64")
                if not os.path.isdir(include_dir):
                    include_dir = os.path.join(mpi_dir, "include")
            self.library_info.update(
                define_macros=define_macros,
                include_dirs=[include_dir],
                libraries=[library],
                library_dirs=[library_dir],
            )
            self.section = name
            self.filename = [mpi_dir]
            return True
        return None

    def setup_compiler_info(self, options, environ):
        def find_exe(cmd, path=None):
            if not cmd:
                return None
            parts = shlex.split(cmd)
            exe, args = parts[0], parts[1:]
            if not os.path.isabs(exe) and path:
                exe = os.path.basename(exe)
            exe = shutil.which(exe, path=path)
            if not exe:
                return None
            return " ".join([exe, *args])

        COMPILERS = (
            ("mpicc", ["mpicc"]),
            ("mpicxx", ["mpicxx", "mpic++", "mpiCC"]),
            ("mpild", []),
        )
        #
        compiler_info = {}
        PATH = environ.get("PATH", "")
        for name, _ in COMPILERS:
            cmd = (
                environ.get(f"MPI4PY_BUILD_{name.upper()}")
                or environ.get(name.upper())
                or getattr(options, name, None)
                or self.compiler_info.get(name)
                or None
            )
            if cmd:
                exe = find_exe(cmd, path=PATH)
                if exe:
                    path = os.path.dirname(exe)
                    PATH = path + os.path.pathsep + PATH
                    compiler_info[name] = exe
                else:
                    self.log.warn("warning: %s='%s' not found", name, cmd)
        #
        if not self and not compiler_info:
            for name, candidates in COMPILERS:
                for cmd in candidates:
                    cmd = find_exe(cmd)
                    if cmd:
                        compiler_info[name] = cmd
                        break
        #
        self.compiler_info.update(compiler_info)

    def load(self, filename="mpi.cfg", section="mpi"):
        if isinstance(filename, str):
            filenames = filename.split(os.path.pathsep)
        else:
            filenames = list(filename)
        if isinstance(section, str):
            sections = section.split(",")
        else:
            sections = list(section)
        #
        parser = ConfigParser(dict_type=OrderedDict)
        try:
            read_ok = parser.read(filenames)
        except ConfigParserError:
            read_ok = None
        if read_ok is None:
            self.log.error(
                "error: parsing configuration file/s '%s'",
                os.path.pathsep.join(filenames),
            )
            return None
        for section in sections:
            if parser.has_section(section):
                break
            section = None
        if not section:
            self.log.error(
                "error: section/s '%s' not found in file/s '%s'",
                ",".join(sections),
                os.path.pathsep.join(filenames),
            )
            return None
        parser_items = list(parser.items(section, vars=None))
        #
        compiler_info = type(self.compiler_info)()
        for option, value in parser_items:
            if option in self.compiler_info:
                compiler_info[option] = value
        #
        pathsep = os.path.pathsep
        expanduser = os.path.expanduser
        expandvars = os.path.expandvars
        library_info = type(self.library_info)()
        for k, v in parser_items:
            if k in (
                "define_macros",
                "undef_macros",
            ):
                macros = [e.strip() for e in v.split(",")]
                if k == "define_macros":
                    for i, m in enumerate(macros):
                        name, _, value = m.partition("=")
                        macros[i] = (name, value or None)
                library_info[k] = macros
            elif k in (
                "include_dirs",
                "library_dirs",
                "rpath",
                "runtime_dirs",
                "runtime_library_dirs",
            ):
                if k in ("rpath", "runtime_dirs"):
                    k = "runtime_library_dirs"
                pathlist = [p.strip() for p in v.split(pathsep)]
                library_info[k] = [
                    expanduser(expandvars(p)) for p in pathlist if p
                ]
            elif k == "libraries":
                library_info[k] = [e.strip() for e in shlex.split(v)]
            elif k in (
                "extra_compile_args",
                "extra_link_args",
            ):
                library_info[k] = shlex.split(v)
            elif k == "extra_objects":
                library_info[k] = [
                    expanduser(expandvars(e)) for e in shlex.split(v)
                ]
            elif hasattr(self, k):
                library_info[k] = v.strip()
        #
        self.section = section
        self.filename = read_ok
        self.compiler_info.update(compiler_info)
        self.library_info.update(library_info)
        return compiler_info, library_info, section, read_ok

    def dump(self, filename=None, section="mpi"):
        # prepare configuration values
        compiler_info = self.compiler_info.copy()
        library_info = self.library_info.copy()
        for k in library_info:
            if k in (
                "define_macros",
                "undef_macros",
            ):
                macros = library_info[k]
                if k == "define_macros":
                    for i, (m, v) in enumerate(macros):
                        if v is None:
                            macros[i] = m
                        else:
                            macros[i] = f"{m}={v}"
                library_info[k] = ",".join(macros)
            elif k in (
                "include_dirs",
                "library_dirs",
                "runtime_library_dirs",
            ):
                library_info[k] = os.path.pathsep.join(library_info[k])
            elif isinstance(library_info[k], list):
                library_info[k] = " ".join(library_info[k])
        # fill configuration parser
        parser = ConfigParser(dict_type=OrderedDict)
        parser.add_section(section)
        for option, value in compiler_info.items():
            if not value:
                continue
            parser.set(section, option, value)
        for option, value in library_info.items():
            if not value:
                continue
            parser.set(section, option, value)
        # save configuration file
        if filename is None:
            parser.write(sys.stdout)
        elif hasattr(filename, "write"):
            parser.write(filename)
        elif isinstance(filename, str):
            with open(filename, "w", encoding="utf-8") as f:
                parser.write(f)
        return parser


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.color = True  # Python 3.14+
    parser.add_argument("--mpi")
    parser.add_argument("--mpicc")
    parser.add_argument("--mpicxx")
    parser.add_argument("--mpild")
    opts = parser.parse_args()

    cfg = Config()
    cfg.setup(opts)
    cfg.dump()



# -----------------------------------------------------------------------------

cdef inline str combinername(int combiner):
    if combiner == MPI_COMBINER_NAMED          : return 'NAMED'
    if combiner == MPI_COMBINER_DUP            : return 'DUP'
    if combiner == MPI_COMBINER_CONTIGUOUS     : return 'CONTIGUOUS'
    if combiner == MPI_COMBINER_VECTOR         : return 'VECTOR'
    if combiner == MPI_COMBINER_HVECTOR        : return 'HVECTOR'
    if combiner == MPI_COMBINER_INDEXED        : return 'INDEXED'
    if combiner == MPI_COMBINER_HINDEXED       : return 'HINDEXED'
    if combiner == MPI_COMBINER_INDEXED_BLOCK  : return 'INDEXED_BLOCK'
    if combiner == MPI_COMBINER_HINDEXED_BLOCK : return 'HINDEXED_BLOCK'
    if combiner == MPI_COMBINER_STRUCT         : return 'STRUCT'
    if combiner == MPI_COMBINER_SUBARRAY       : return 'SUBARRAY'
    if combiner == MPI_COMBINER_DARRAY         : return 'DARRAY'
    if combiner == MPI_COMBINER_RESIZED        : return 'RESIZED'
    if combiner == MPI_COMBINER_VALUE_INDEX    : return 'VALUE_INDEX'
    if combiner == MPI_COMBINER_F90_INTEGER    : return 'F90_INTEGER'
    if combiner == MPI_COMBINER_F90_REAL       : return 'F90_REAL'
    if combiner == MPI_COMBINER_F90_COMPLEX    : return 'F90_COMPLEX'
    raise ValueError(f"unknown combiner value {combiner}")  # ~> unreachable

cdef inline list makelist(integral_t *p, MPI_Count start, MPI_Count last):
    return [p[i] for i from start <= i <= last]

cdef inline tuple datatype_decode(
    Datatype self,
    bint mark,
):
    # get the datatype envelope
    cdef int combiner = MPI_UNDEFINED
    cdef MPI_Count ni = 0, na = 0, nc = 0, nd = 0
    CHKERR( MPI_Type_get_envelope_c(
        self.ob_mpi, &ni, &na, &nc, &nd, &combiner) )
    # return self immediately for named datatypes
    if combiner == MPI_COMBINER_NAMED:
        return (self, combinername(combiner), {})
    # get the datatype contents
    cdef int *i = NULL
    cdef MPI_Aint *a = NULL
    cdef MPI_Count *c = NULL
    cdef MPI_Datatype *d = NULL
    cdef unused1 = allocate(ni, sizeof(int), &i)
    cdef unused2 = allocate(na, sizeof(MPI_Aint), &a)
    cdef unused3 = allocate(nc, sizeof(MPI_Count), &c)
    cdef unused4 = allocate(nd, sizeof(MPI_Datatype), &d)
    CHKERR( MPI_Type_get_contents_c(
        self.ob_mpi, ni, na, nc, nd, i, a, c, d) )
    # process datatypes in advance
    cdef Datatype oldtype = __DATATYPE_NULL__
    cdef dict params = {}
    cdef list datatypes = []
    if nd == 1:
        oldtype = ref_Datatype(d[0])
        datatypes = [oldtype]
    elif nd > 1:
        datatypes = [ref_Datatype(d[k]) for k in range(nd)]

    # dispatch depending on the combiner value
    cdef int use_count = 1 if (nc > 0) else 0
    cdef MPI_Count s1, e1, s2, e2, s3, e3, s4, e4
    cdef object count, blklen, stride, displs
    cdef object sizes, subsizes, starts, order
    cdef object lbound, extent
    if combiner == MPI_COMBINER_DUP:
        params = {}
    elif combiner == MPI_COMBINER_CONTIGUOUS:
        count = c[0] if use_count else i[0]
        params = {
            ('count') : count,
        }
    elif combiner == MPI_COMBINER_VECTOR:
        count  = c[0] if use_count else i[0]
        blklen = c[1] if use_count else i[1]
        stride = c[2] if use_count else i[2]
        params = {
            ('count')       : count,
            ('blocklength') : blklen,
            ('stride')      : stride,
        }
    elif combiner == MPI_COMBINER_HVECTOR:
        count  = c[0] if use_count else i[0]
        blklen = c[1] if use_count else i[1]
        stride = c[2] if use_count else a[0]
        params = {
            ('count')       : count ,
            ('blocklength') : blklen,
            ('stride')      : stride,
        }
    elif combiner == MPI_COMBINER_INDEXED:
        if use_count:
            s1, e1 =      1,   c[0]
            s2, e2 = c[0]+1, 2*c[0]
            blklen = makelist(c, s1, e1)
            displs = makelist(c, s2, e2)
        else:
            s1, e1 =      1,   i[0]       # ~> uncovered
            s2, e2 = i[0]+1, 2*i[0]       # ~> uncovered
            blklen = makelist(i, s1, e1)  # ~> uncovered
            displs = makelist(i, s2, e2)  # ~> uncovered
        params = {
            ('blocklengths')  : blklen,
            ('displacements') : displs,
        }
    elif combiner == MPI_COMBINER_HINDEXED:
        if use_count:
            s1, e1 =      1,   c[0]
            s2, e2 = c[0]+1, 2*c[0]
            blklen = makelist(c, s1, e1)
            displs = makelist(c, s2, e2)
        else:
            s1, e1 = 1, i[0]              # ~> uncovered
            s2, e2 = 0, i[0]-1            # ~> uncovered
            blklen = makelist(i, s1, e1)  # ~> uncovered
            displs = makelist(a, s2, e2)  # ~> uncovered
        params = {
            ('blocklengths')  : blklen,
            ('displacements') : displs,
        }
    elif combiner == MPI_COMBINER_INDEXED_BLOCK:
        if use_count:
            s2, e2 = 2, c[0]+1
            blklen = c[1]
            displs = makelist(c, s2, e2)
        else:
            s2, e2 = 2, i[0]+1            # ~> uncovered
            blklen = i[1]                 # ~> uncovered
            displs = makelist(i, s2, e2)  # ~> uncovered
        params = {
            ('blocklength')   : blklen,
            ('displacements') : displs,
        }
    elif combiner == MPI_COMBINER_HINDEXED_BLOCK:
        if use_count:
            s2, e2 = 2, c[0]+1
            blklen = c[1]
            displs = makelist(c, s2, e2)
        else:
            s2, e2 = 0, i[0]-1            # ~> uncovered
            blklen = i[1]                 # ~> uncovered
            displs = makelist(a, s2, e2)  # ~> uncovered
        params = {
            ('blocklength')   : blklen,
            ('displacements') : displs,
        }
    elif combiner == MPI_COMBINER_STRUCT:
        if use_count:
            s1, e1 =      1,   c[0]
            s2, e2 = c[0]+1, 2*c[0]
            blklen = makelist(c, s1, e1)
            displs = makelist(c, s2, e2)
        else:
            s1, e1 = 1, i[0]              # ~> uncovered
            s2, e2 = 0, i[0]-1            # ~> uncovered
            blklen = makelist(i, s1, e1)  # ~> uncovered
            displs = makelist(a, s2, e2)  # ~> uncovered
        params = {
            ('blocklengths')  : blklen,
            ('displacements') : displs,
            ('datatypes')     : datatypes,
        }
    elif combiner == MPI_COMBINER_SUBARRAY:
        if use_count:
            s1, e1 = 0*i[0], 1*i[0]-1
            s2, e2 = 1*i[0], 2*i[0]-1
            s3, e3 = 2*i[0], 3*i[0]-1
            sizes    = makelist(c, s1, e1)
            subsizes = makelist(c, s2, e2)
            starts   = makelist(c, s3, e3)
            order    = i[1]
        else:
            s1, e1 = 0*i[0]+1, 1*i[0]       # ~> uncovered
            s2, e2 = 1*i[0]+1, 2*i[0]       # ~> uncovered
            s3, e3 = 2*i[0]+1, 3*i[0]       # ~> uncovered
            sizes    = makelist(i, s1, e1)  # ~> uncovered
            subsizes = makelist(i, s2, e2)  # ~> uncovered
            starts   = makelist(i, s3, e3)  # ~> uncovered
            order    = i[3*i[0]+1]          # ~> uncovered
        params = {
            ('sizes')    : sizes,
            ('subsizes') : subsizes,
            ('starts')   : starts,
            ('order')    : order,
        }
    elif combiner == MPI_COMBINER_DARRAY:
        if use_count:
            s1, e1 = 0*i[2]+0, 1*i[2]+0-1
            s2, e2 = 0*i[2]+3, 1*i[2]+3-1
            s3, e3 = 1*i[2]+3, 2*i[2]+3-1
            s4, e4 = 2*i[2]+3, 3*i[2]+3-1
            sizes = makelist(c, s1, e1)
            order = i[3*i[2]+3]
        else:
            s1, e1 = 0*i[2]+3, 1*i[2]+3-1  # ~> uncovered
            s2, e2 = 1*i[2]+3, 2*i[2]+3-1  # ~> uncovered
            s3, e3 = 2*i[2]+3, 3*i[2]+3-1  # ~> uncovered
            s4, e4 = 3*i[2]+3, 4*i[2]+3-1  # ~> uncovered
            sizes = makelist(i, s1, e1)    # ~> uncovered
            order = i[4*i[2]+3]            # ~> uncovered
        params = {
            ('size')     : i[0],
            ('rank')     : i[1],
            ('gsizes')   : sizes,
            ('distribs') : makelist(i, s2, e2),
            ('dargs')    : makelist(i, s3, e3),
            ('psizes')   : makelist(i, s4, e4),
            ('order')    : order,
        }
    elif combiner == MPI_COMBINER_RESIZED:
        lbound = c[0] if use_count else a[0]
        extent = c[1] if use_count else a[1]
        params = {
            ('lb')     : lbound,
            ('extent') : extent,
        }
    elif combiner == MPI_COMBINER_VALUE_INDEX:
        params = {
            ('value') : datatypes[0],  # ~> MPI-4.1
            ('index') : datatypes[1],  # ~> MPI-4.1
        }
    elif combiner == MPI_COMBINER_F90_INTEGER:
        params = {
            ('r') : i[0],
        }
    elif combiner == MPI_COMBINER_F90_REAL:
        params = {
            ('p') : i[0],
            ('r') : i[1],
        }
    elif combiner == MPI_COMBINER_F90_COMPLEX:
        params = {
            ('p') : i[0],
            ('r') : i[1],
        }
    if mark:
        datatype_visit(marktemp, datatypes)
    return (oldtype, combinername(combiner), params)


cdef inline Datatype datatype_create(
    Datatype datatype,
    str  combiner,
    dict params,
    bint free,
):
    cdef object factory
    cdef Datatype newtype
    cdef list datatypes
    if 'datatypes' in params:
        datatypes = params['datatypes']
    elif 'value' in params and 'index' in params:
        datatypes = list(params.values())  # ~> MPI-4.1
    else:
        datatypes = [datatype]
    try:
        combiner = combiner.lower()
        factory = getattr(datatype, f'Create_{combiner}')
        newtype = factory(**params).Commit()
    finally:
        if free:
            datatype_visit(freetemp, datatypes)
    return newtype


cdef inline int datatype_visit(
    int (*visit)(Datatype) except -1,
    list datatypes,
) except -1:
    cdef Datatype datatype
    for datatype in datatypes:
        visit(datatype)


def _datatype_create(
    Datatype datatype: Datatype,
    str combiner: str,
    dict params: dict[str, Any],
    bint free: bool = False,
) -> Datatype:
    """
    Create datatype from base datatype, combiner name, and parameters.
    """
    return datatype_create(datatype, combiner, params, free)


def _datatype_decode(
    Datatype datatype: Datatype,
    bint mark: bool = False,
) -> tuple[Datatype, str, dict[str, Any]]:
    """
    Decode datatype to base datatype, combiner name, and parameters.
    """
    return datatype_decode(datatype, mark)  # ~> TODO


# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int is_list  "PyList_Check"  (object)
    int is_tuple "PyTuple_Check" (object)

cdef inline int is_buffer(object ob) noexcept:
    return PyObject_CheckBuffer(ob)

cdef inline int is_dlpack_buffer(object ob) noexcept:
    return Py_CheckDLPackBuffer(ob)

cdef inline int is_cai_buffer(object ob) noexcept:
    return Py_CheckCAIBuffer(ob)

cdef inline int is_datatype(object ob) except -1:
    if isinstance(ob, Datatype): return 1
    if isinstance(ob, str): return 1
    return 0

cdef inline bint is_constant(object obj, object CONST):
    if PYPY: return type(obj) is type(CONST) and obj == CONST
    return obj is CONST

# -----------------------------------------------------------------------------

cdef inline const char *getformat(const char format[]) except NULL:
    cdef char byteorder = format[0]
    if byteorder == c'@':  # native
        return format + 1                            # ~> uncovered
    if byteorder == c'<':  # little-endian
        if not is_little_endian():
            raise BufferError(                       # ~> big-endian
                f"format string {pystr(format)!r} "  # ~> big-endian
                f"with non-native byte order")       # ~> big-endian
        return format + 1                            # ~> little-endian
    if byteorder == c'>':  # big-endian
        if not is_big_endian():
            raise BufferError(                       # ~> little-endian
                f"format string {pystr(format)!r} "  # ~> little-endian
                f"with non-native byte order")       # ~> little-endian
        return format + 1                            # ~> big-endian
    # passthrough
    return format

cdef inline Datatype lookup_datatype(object key):
    try:
        return <Datatype?> TypeDict[key]
    except KeyError:
        raise KeyError(f"cannot map {key!r} to MPI datatype")

cdef inline Datatype asdatatype(object datatype):
    if isinstance(datatype, Datatype):
        return <Datatype> datatype
    return lookup_datatype(datatype)

cdef inline Datatype getdatatype(const char format[]):
    if format == BYTE_FMT: return __BYTE__
    return lookup_datatype(pystr(getformat(format)))


@cython.final
@cython.internal
cdef class _p_message:
    cdef buffer buf
    cdef object count
    cdef object displ
    cdef Datatype type

cdef _p_message message_basic(object o_buf,
                              object o_type,
                              bint readonly,
                              #
                              void        **baddr,
                              MPI_Count    *bsize,
                              MPI_Datatype *btype,
                              ):
    cdef _p_message m = _p_message.__new__(_p_message)
    # special-case for BOTTOM or None,
    # an explicit MPI datatype is required
    if is_BOTTOM(o_buf):
        m.buf = newbuffer()
        m.type = asdatatype(o_type)
        baddr[0] = MPI_BOTTOM
        bsize[0] = 0
        btype[0] = m.type.ob_mpi
        return m
    # get message buffer
    cdef bint fmt = (o_type is None)
    m.buf = getbuffer(o_buf, readonly, fmt)
    # get message datatype
    if o_type is not None:
        m.type = asdatatype(o_type)
    else:
        m.type = getdatatype(m.buf.view.format)
    # return collected message data
    baddr[0] = <void*> m.buf.view.buf
    bsize[0] = <MPI_Aint> m.buf.view.len
    btype[0] = m.type.ob_mpi
    return m

cdef _p_message message_simple(object msg,
                               int readonly,
                               int rank,
                               int blocks,
                               #
                               void         **_addr,
                               MPI_Count    *_count,
                               MPI_Datatype *_type,
                               ):
    # special-case PROC_NULL target rank
    if rank == MPI_PROC_NULL:
        _addr[0]  = NULL
        _count[0] = 0
        _type[0]  = MPI_BYTE
        return None
    # unpack message list/tuple
    cdef Py_ssize_t nargs = 0
    cdef object o_buf   = None
    cdef object o_count = None
    cdef object o_displ = None
    cdef object o_type  = None
    if is_buffer(msg):
        o_buf = msg
    elif is_list(msg) or is_tuple(msg):
        nargs = len(msg)
        if nargs == 2:
            (o_buf, o_count) = msg
            if is_datatype(o_count):
                (o_count, o_type) = None, o_count
            elif is_tuple(o_count) or is_list(o_count):
                (o_count, o_displ) = o_count
        elif nargs == 3:
            (o_buf, o_count, o_type) = msg
            if is_tuple(o_count) or is_list(o_count):
                (o_count, o_displ) = o_count
        elif nargs == 4:
            (o_buf, o_count, o_displ, o_type) = msg
        else:
            raise ValueError("message: expecting 2 to 4 items")
    elif is_dlpack_buffer(msg):
        o_buf = msg
    elif is_cai_buffer(msg):
        o_buf = msg
    else:
        raise TypeError("message: expecting buffer or list/tuple")
    # buffer: address, length, and datatype
    cdef void *baddr = NULL
    cdef MPI_Count bsize = 0
    cdef MPI_Datatype btype = MPI_DATATYPE_NULL
    cdef _p_message m = message_basic(o_buf, o_type, readonly,
                                      &baddr, &bsize, &btype)
    # buffer: count and displacement
    cdef MPI_Count count = 0  # number of datatype entries
    cdef MPI_Aint  displ = 0  # from base buffer, in datatype entries
    cdef MPI_Count extent = 0, lb = 0  # datatype extent
    cdef MPI_Count length = bsize      # in bytes
    cdef MPI_Aint  offset = 0          # from base buffer, in bytes
    if o_displ is not None:
        displ = <MPI_Aint> PyNumber_Index(o_displ)
        if displ < 0:
            raise ValueError(
                f"message: negative displacement {displ}")
        if displ > 0:
            if btype == MPI_DATATYPE_NULL:
                raise ValueError(
                    f"message: cannot handle displacement, "
                    f"datatype is null")
            CHKERR( MPI_Type_get_extent_c(btype, &lb, &extent) )
            if extent <= 0:
                raise ValueError(
                    f"message: cannot handle displacement, "
                    f"datatype extent {extent} (lb:{lb}, ub:{lb+extent})")
            if displ*extent > length:
                raise ValueError(
                    f"message: displacement {displ} out of bounds, "
                    f"number of datatype entries {length//extent}")
            offset = <MPI_Aint> (displ * extent)
            length -= offset
    if o_count is not None:
        count = <MPI_Count> PyNumber_Index(o_count)
        if count < 0:
            raise ValueError(
                f"message: negative count {count}")
        if count > 0 and o_buf is None:
            raise ValueError(
                f"message: buffer is None but count is {count}")
    elif length > 0:
        if extent == 0:
            if btype == MPI_DATATYPE_NULL:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"datatype is null")
            CHKERR( MPI_Type_get_extent_c(btype, &lb, &extent) )
            if extent <= 0:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"datatype extent {extent} (lb:{lb}, ub:{lb+extent})")
        if (length % extent) != 0:
            raise ValueError(
                f"message: cannot infer count, "
                f"buffer length {length} is not a multiple of "
                f"datatype extent {extent} (lb:{lb}, ub:{lb+extent})")
        if blocks < 2:
            count = length // extent
        else:
            if ((length // extent) % blocks) != 0:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"number of entries {length//extent} is not a multiple of "
                    f"required number of blocks {blocks}")
            count = (length // extent) // blocks
    # return collected message data
    m.count = o_count if o_count is not None else count
    m.displ = o_displ if o_displ is not None else displ
    _addr[0]  = <void*>(<char*>baddr + offset)
    _count[0] = count
    _type[0]  = btype
    return m

cdef _p_message message_vector(object msg,
                               int readonly,
                               int rank,
                               int blocks,
                               #
                               void        **_addr,
                               MPI_Count   **_counts,
                               MPI_Aint    **_displs,
                               MPI_Datatype *_type,
                               ):
    # special-case PROC_NULL target rank
    if rank == MPI_PROC_NULL:
        _addr[0]   = NULL
        _counts[0] = NULL
        _displs[0] = NULL
        _type[0]   = MPI_BYTE
        return None
    # unpack message list/tuple
    cdef Py_ssize_t nargs = 0
    cdef object o_buf    = None
    cdef object o_counts = None
    cdef object o_displs = None
    cdef object o_type   = None
    if is_buffer(msg):
        o_buf = msg
    elif is_list(msg) or is_tuple(msg):
        nargs = len(msg)
        if nargs == 2:
            (o_buf, o_counts) = msg
            if is_datatype(o_counts):
                (o_counts, o_type) = None, o_counts
            elif is_tuple(o_counts):
                (o_counts, o_displs) = o_counts
        elif nargs == 3:
            (o_buf, o_counts, o_type) = msg
            if is_tuple(o_counts):
                (o_counts, o_displs) = o_counts
        elif nargs == 4:
            (o_buf, o_counts, o_displs, o_type) = msg
        else:
            raise ValueError("message: expecting 2 to 4 items")
    elif is_dlpack_buffer(msg):
        o_buf = msg
    elif is_cai_buffer(msg):
        o_buf = msg
    else:
        raise TypeError("message: expecting buffer or list/tuple")
    # buffer: address, length, and datatype
    cdef void *baddr = NULL
    cdef MPI_Count bsize = 0
    cdef MPI_Datatype btype = MPI_DATATYPE_NULL
    cdef _p_message m = message_basic(o_buf, o_type, readonly,
                                      &baddr, &bsize, &btype)
    # counts and displacements
    cdef int i
    cdef MPI_Count *counts = NULL
    cdef MPI_Aint  *displs = NULL
    cdef MPI_Count extent=0, lb=0
    cdef MPI_Aint  avalue=0
    cdef MPI_Count cvalue=0, csize=0
    if o_counts is None:
        if bsize > 0:
            if btype == MPI_DATATYPE_NULL:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"datatype is null")
            CHKERR( MPI_Type_get_extent_c(btype, &lb, &extent) )
            if extent <= 0:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"datatype extent {extent} (lb:{lb}, ub:{lb+extent})")
            if (bsize % extent) != 0:
                raise ValueError(
                    f"message: cannot infer count, "
                    f"buffer length {bsize} is not a multiple of "
                    f"datatype extent {extent} (lb:{lb}, ub:{lb+extent})")
            csize = bsize // extent
        o_counts = newarray(blocks, &counts)
        for i in range(blocks):
            cvalue = (csize // blocks) + (csize % blocks > i)
            counts[i] = cvalue
    elif is_integral(o_counts):
        cvalue = <MPI_Count> PyNumber_Index(o_counts)
        o_counts = newarray(blocks, &counts)
        for i in range(blocks):
            counts[i] = cvalue
    else:
        o_counts = chkarray(o_counts, blocks, &counts)
    if o_displs is None:  # contiguous
        avalue = 0
        o_displs = newarray(blocks, &displs)
        for i in range(blocks):
            displs[i] = avalue
            avalue += <MPI_Aint> counts[i]
    elif is_integral(o_displs):  # strided
        avalue = <MPI_Aint> PyNumber_Index(o_displs)
        o_displs = newarray(blocks, &displs)
        for i in range(blocks):
            displs[i] = avalue * i
    else:  # general
        o_displs = chkarray(o_displs, blocks, &displs)
    # return collected message data
    m.count = o_counts
    m.displ = o_displs
    _addr[0]   = baddr
    _counts[0] = counts
    _displs[0] = displs
    _type[0]   = btype
    return m

cdef tuple message_vector_w(object msg,
                            int readonly,
                            int blocks,
                            #
                            void         **_addr,
                            MPI_Count    **_counts,
                            MPI_Aint     **_displs,
                            MPI_Datatype **_types,
                            ):
    cdef Py_ssize_t nargs = len(msg)
    cdef object o_buffer, o_counts, o_displs, o_types
    if nargs == 2:
        o_buffer, o_types = msg
        o_counts = o_displs = None
    elif nargs == 3:
        o_buffer, (o_counts, o_displs), o_types = msg
    elif nargs == 4:
        o_buffer,  o_counts, o_displs,  o_types = msg
    else:
        raise ValueError("message: expecting 2 to 4 items")
    if is_BOTTOM(o_buffer):
        if o_counts is None:
            raise ValueError("message: BOTTOM requires counts")
        if o_displs is None:
            raise ValueError("message: BOTTOM requires displs")
        _addr[0] = MPI_BOTTOM
    elif readonly:
        o_buffer = asbuffer_r(o_buffer, _addr, NULL)
    else:
        o_buffer = asbuffer_w(o_buffer, _addr, NULL)
    if o_counts is None and o_displs is None:
        o_counts = newarray(blocks, _counts)
        o_displs = newarray(blocks, _displs)
        for i in range(blocks):
            _counts[0][i] = 1
            _displs[0][i] = 0
    else:
        o_counts = chkarray(o_counts, blocks, _counts)
        o_displs = chkarray(o_displs, blocks, _displs)
    o_types  = asarray_Datatype(o_types, blocks, _types)
    return (o_buffer, o_counts, o_displs, o_types)

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_msg_p2p:

    # raw C-side arguments
    cdef void         *buf
    cdef MPI_Count    count
    cdef MPI_Datatype dtype
    # python-side argument
    cdef object _msg

    def __cinit__(self):
        self.buf   = NULL
        self.count = 0
        self.dtype = MPI_DATATYPE_NULL

    cdef int for_send(self, object msg, int rank, int parts) except -1:
        self._msg = message_simple(
            msg, 1,  # readonly
            rank, parts,
            &self.buf,
            &self.count,
            &self.dtype,
        )
        return 0

    cdef int for_recv(self, object msg, int rank, int parts) except -1:
        self._msg = message_simple(
            msg, 0,  # writable
            rank, parts,
            &self.buf,
            &self.count,
            &self.dtype,
        )
        return 0


cdef inline _p_msg_p2p message_p2p_send(object sendbuf, int dest):
    cdef _p_msg_p2p msg = _p_msg_p2p.__new__(_p_msg_p2p)
    msg.for_send(sendbuf, dest, 1)
    return msg

cdef inline _p_msg_p2p message_p2p_recv(object recvbuf, int source):
    cdef _p_msg_p2p msg = _p_msg_p2p.__new__(_p_msg_p2p)
    msg.for_recv(recvbuf, source, 1)
    return msg

cdef inline _p_msg_p2p message_p2p_psend(object sendbuf, int dest, int parts):
    cdef _p_msg_p2p msg = _p_msg_p2p.__new__(_p_msg_p2p)
    msg.for_send(sendbuf, dest, parts)
    return msg

cdef inline _p_msg_p2p message_p2p_precv(object recvbuf, int source, int parts):
    cdef _p_msg_p2p msg = _p_msg_p2p.__new__(_p_msg_p2p)
    msg.for_recv(recvbuf, source, parts)
    return msg

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_msg_cco:

    # raw C-side arguments
    cdef void *sbuf, *rbuf
    cdef MPI_Count  scount, rcount
    cdef MPI_Count *scounts, *rcounts
    cdef MPI_Aint  *sdispls, *rdispls
    cdef MPI_Datatype stype, rtype
    # python-side arguments
    cdef object _smsg, _rmsg
    cdef object _rcnt

    def __cinit__(self):
        self.sbuf    = self.rbuf    = NULL
        self.scount  = self.rcount  = 0
        self.scounts = self.rcounts = NULL
        self.sdispls = self.rdispls = NULL
        self.stype   = self.rtype   = MPI_DATATYPE_NULL

    # Collective Communication Operations
    # -----------------------------------

    # sendbuf arguments
    cdef int for_cco_send(
        self, bint VECTOR, object amsg,
        int rank, int blocks,
    ) except -1:
        cdef bint readonly = 1
        if not VECTOR:  # block variant
            self._smsg = message_simple(
                amsg, readonly, rank, blocks,
                &self.sbuf, &self.scount, &self.stype)
        else:  # vector variant
            self._smsg = message_vector(
                amsg, readonly, rank, blocks,
                &self.sbuf, &self.scounts,
                &self.sdispls, &self.stype)
        return 0

    # recvbuf arguments
    cdef int for_cco_recv(
        self, bint VECTOR, object amsg,
        int rank, int blocks,
    ) except -1:
        cdef bint readonly = 0
        if not VECTOR:  # block variant
            self._rmsg = message_simple(
                amsg, readonly, rank, blocks,
                &self.rbuf, &self.rcount, &self.rtype)
        else:  # vector variant
            self._rmsg = message_vector(
                amsg, readonly, rank, blocks,
                &self.rbuf, &self.rcounts,
                &self.rdispls, &self.rtype)
        return 0

    # bcast
    cdef int for_bcast(
        self, object msg,
        int root, MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, rank=0, sending=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_rank(comm, &rank) )
            if root == rank:
                self.for_cco_send(0, msg, root, 0)
                sending = 1
            else:
                self.for_cco_recv(0, msg, root, 0)
                sending = 0
        else:  # inter-communication
            if root == MPI_ROOT or root == MPI_PROC_NULL:
                self.for_cco_send(0, msg, root, 0)
                sending = 1
            else:
                self.for_cco_recv(0, msg, root, 0)
                sending = 0
        if sending:
            self.rbuf   = self.sbuf
            self.rcount = self.scount
            self.rtype  = self.stype
        else:
            self.sbuf   = self.rbuf
            self.scount = self.rcount
            self.stype  = self.rtype
        return 0

    # gather/gatherv
    cdef int for_gather(
        self, int v, object smsg, object rmsg,
        int root, MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0, rank=0, null=MPI_PROC_NULL
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_size(comm, &size) )
            CHKERR( MPI_Comm_rank(comm, &rank) )
            if root == rank:
                self.for_cco_recv(v, rmsg, root, size)
                if is_IN_PLACE(smsg):
                    self.sbuf = MPI_IN_PLACE
                else:
                    self.for_cco_send(0, smsg, 0, 0)
            else:
                self.for_cco_recv(v, rmsg, null, size)
                self.for_cco_send(0, smsg, root, 0)
        else:  # inter-communication
            CHKERR( MPI_Comm_remote_size(comm, &size) )
            if root == MPI_ROOT or root == MPI_PROC_NULL:
                self.for_cco_recv(v, rmsg, root, size)
                self.for_cco_send(0, smsg, null, 0)
            else:
                self.for_cco_recv(v, rmsg, null, size)
                self.for_cco_send(0, smsg, root, 0)
        return 0

    # scatter/scatterv
    cdef int for_scatter(
        self, int v,
        object smsg, object rmsg,
        int root, MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0, rank=0, null=MPI_PROC_NULL
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_size(comm, &size) )
            CHKERR( MPI_Comm_rank(comm, &rank) )
            if root == rank:
                self.for_cco_send(v, smsg, root, size)
                if is_IN_PLACE(rmsg):
                    self.rbuf = MPI_IN_PLACE
                else:
                    self.for_cco_recv(0, rmsg, root, 0)
            else:
                self.for_cco_send(v, smsg, null, size)
                self.for_cco_recv(0, rmsg, root, 0)
        else:  # inter-communication
            CHKERR( MPI_Comm_remote_size(comm, &size) )
            if root == MPI_ROOT or root == MPI_PROC_NULL:
                self.for_cco_send(v, smsg, root, size)
                self.for_cco_recv(0, rmsg, null,  0)
            else:
                self.for_cco_send(v, smsg, null, size)
                self.for_cco_recv(0, rmsg, root, 0)
        return 0

    # allgather/allgatherv
    cdef int for_allgather(
        self, int v, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_size(comm, &size) )
        else:  # inter-communication
            CHKERR( MPI_Comm_remote_size(comm, &size) )
        #
        self.for_cco_recv(v, rmsg, 0, size)
        if not inter and is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cco_send(0, smsg, 0, 0)
        return 0

    # alltoall/alltoallv
    cdef int for_alltoall(
        self, int v, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_size(comm, &size) )
        else:  # inter-communication
            CHKERR( MPI_Comm_remote_size(comm, &size) )
        #
        self.for_cco_recv(v, rmsg, 0, size)
        if not inter and is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cco_send(v, smsg, 0, size)
        return 0

    # Neighbor Collectives
    # --------------------

    # neighbor allgather/allgatherv
    cdef int for_neighbor_allgather(
        self, int v, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int recvsize=0
        comm_neighbors_count(comm, &recvsize, NULL)
        self.for_cco_send(0, smsg, 0, 0)
        self.for_cco_recv(v, rmsg, 0, recvsize)
        return 0

    # neighbor alltoall/alltoallv
    cdef int for_neighbor_alltoall(
        self, int v, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int sendsize=0, recvsize=0
        comm_neighbors_count(comm, &recvsize, &sendsize)
        self.for_cco_send(v, smsg, 0, sendsize)
        self.for_cco_recv(v, rmsg, 0, recvsize)
        return 0

    # Collective Reductions Operations
    # --------------------------------

    # sendbuf
    cdef int for_cro_send(
        self, object amsg, int root,
    ) except -1:
        cdef bint readonly = 1
        self._smsg = message_simple(
            amsg, readonly, root, 0,
            &self.sbuf, &self.scount, &self.stype)
        return 0

    # recvbuf
    cdef int for_cro_recv(
        self, object amsg, int root,
    ) except -1:
        cdef bint readonly = 0
        self._rmsg = message_simple(
            amsg, readonly, root, 0,
            &self.rbuf, &self.rcount, &self.rtype)
        return 0

    # check counts and datatypes
    cdef int chk_cro_args(self) except -1:
        if self.sbuf == MPI_IN_PLACE: return 0
        if self.stype != self.rtype:
            raise ValueError(
                f"mismatch in send and receive MPI datatypes")
        if self.scount != self.rcount:
            raise ValueError(
                f"mismatch in send count {self.scount} "
                f"and receive count {self.rcount}")
        return 0

    cdef int for_reduce(
        self, object smsg, object rmsg,
        int root, MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, rank=0, null=MPI_PROC_NULL
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_rank(comm, &rank) )
            if root == rank:
                self.for_cro_recv(rmsg, root)
                if is_IN_PLACE(smsg):
                    self.sbuf = MPI_IN_PLACE
                else:
                    self.for_cro_send(smsg, root)
                    self.chk_cro_args()
            else:
                self.for_cro_recv(rmsg, null)
                self.for_cro_send(smsg, root)
                self.rcount = self.scount
                self.rtype  = self.stype
        else:  # inter-communication
            if root == MPI_ROOT or root == MPI_PROC_NULL:
                self.for_cro_recv(rmsg, root)
                self.scount = self.rcount
                self.stype  = self.rtype
            else:
                self.for_cro_send(smsg, root)
                self.rcount = self.scount
                self.rtype  = self.stype
        return 0

    cdef int for_allreduce(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        # get send and recv buffers
        self.for_cro_recv(rmsg, 0)
        if not inter and is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cro_send(smsg, 0)
            self.chk_cro_args()
        return 0

    cdef int for_reduce_scatter_block(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        CHKERR( MPI_Comm_size(comm, &size) )
        # get send and recv buffers
        if not inter and is_IN_PLACE(smsg):
            self.for_cco_recv(0, rmsg, 0, size)
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cco_recv(0, rmsg, 0, 0)
            self.for_cco_send(0, smsg, 0, size)
            self.chk_cro_args()
        return 0

    cdef int for_reduce_scatter(
        self, object smsg, object rmsg, object rcnt,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0, rank=MPI_PROC_NULL
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        CHKERR( MPI_Comm_size(comm, &size) )
        CHKERR( MPI_Comm_rank(comm, &rank) )
        # get send and recv buffers
        self.for_cro_recv(rmsg, 0)
        if not inter and is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cro_send(smsg, 0)
        # get receive counts
        if rcnt is None and not inter and self.sbuf != MPI_IN_PLACE:
            self._rcnt = newarray(size, &self.rcounts)
            CHKERR( MPI_Allgather_c(
                &self.rcount, 1, MPI_COUNT,
                self.rcounts, 1, MPI_COUNT, comm) )
        else:
            self._rcnt = chkarray(rcnt, size, &self.rcounts)
        # total sum or receive counts
        cdef MPI_Count sumrcounts=0
        for i in range(size):
            sumrcounts += self.rcounts[i]
        # check counts and datatypes
        if self.sbuf != MPI_IN_PLACE:
            if self.stype != self.rtype:
                raise ValueError(
                    f"mismatch in send and receive MPI datatypes")
            if self.scount != sumrcounts:
                raise ValueError(
                    f"mismatch in send count {self.scount} "
                    f"and sum(counts) {sumrcounts}")
            if self.rcount != self.rcounts[rank]:
                raise ValueError(
                    f"mismatch in receive count {self.rcount} "
                    f"and counts[{rank}] {self.rcounts[rank]}")
        else:
            if self.rcount != sumrcounts:
                raise ValueError(
                    f"mismatch in receive count {self.rcount} "
                    f"and sum(counts) {sumrcounts}")
        return 0

    cdef int for_scan(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        # get send and recv buffers
        self.for_cro_recv(rmsg, 0)
        if is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cro_send(smsg, 0)
            self.chk_cro_args()

    cdef int for_exscan(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        # get send and recv buffers
        self.for_cro_recv(rmsg, 0)
        if is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
        else:
            self.for_cro_send(smsg, 0)
            self.chk_cro_args()


cdef inline _p_msg_cco message_cco():
    cdef _p_msg_cco msg = _p_msg_cco.__new__(_p_msg_cco)
    return msg

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_msg_ccow:

    # raw C-side arguments
    cdef void *sbuf, *rbuf
    cdef MPI_Count *scounts, *rcounts
    cdef MPI_Aint  *sdispls, *rdispls
    cdef MPI_Datatype *stypes, *rtypes
    # python-side arguments
    cdef object _smsg, _rmsg

    def __cinit__(self):
        self.sbuf     = self.rbuf     = NULL
        self.scounts  = self.rcounts  = NULL
        self.sdispls  = self.rdispls  = NULL
        self.stypes   = self.rtypes   = NULL

    # alltoallw
    cdef int for_alltoallw(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int inter=0, size=0
        CHKERR( MPI_Comm_test_inter(comm, &inter) )
        if not inter:  # intra-communication
            CHKERR( MPI_Comm_size(comm, &size) )
        else:  # inter-communication
            CHKERR( MPI_Comm_remote_size(comm, &size) )
        #
        self._rmsg = message_vector_w(
            rmsg, 0, size,
            &self.rbuf, &self.rcounts,
            &self.rdispls, &self.rtypes)
        if not inter and is_IN_PLACE(smsg):
            self.sbuf = MPI_IN_PLACE
            return 0
        self._smsg = message_vector_w(
            smsg, 1, size,
            &self.sbuf, &self.scounts,
            &self.sdispls, &self.stypes)
        return 0

    # neighbor alltoallw
    cdef int for_neighbor_alltoallw(
        self, object smsg, object rmsg,
        MPI_Comm comm,
    ) except -1:
        if comm == MPI_COMM_NULL: return 0
        cdef int sendsize=0, recvsize=0
        comm_neighbors_count(comm, &recvsize, &sendsize)
        self._rmsg = message_vector_w(
            rmsg, 0, recvsize,
            &self.rbuf, &self.rcounts,
            &self.rdispls, &self.rtypes)
        self._smsg = message_vector_w(
            smsg, 1, sendsize,
            &self.sbuf, &self.scounts,
            &self.sdispls, &self.stypes)
        return 0


cdef inline _p_msg_ccow message_ccow():
    cdef _p_msg_ccow msg = _p_msg_ccow.__new__(_p_msg_ccow)
    return msg

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_msg_rma:

    # raw origin arguments
    cdef void*        oaddr
    cdef MPI_Count    ocount
    cdef MPI_Datatype otype
    # raw compare arguments
    cdef void*        caddr
    cdef MPI_Count    ccount
    cdef MPI_Datatype ctype
    # raw result arguments
    cdef void*        raddr
    cdef MPI_Count    rcount
    cdef MPI_Datatype rtype
    # raw target arguments
    cdef MPI_Aint     tdisp
    cdef MPI_Count    tcount
    cdef MPI_Datatype ttype
    # python-side arguments
    cdef object _origin
    cdef object _compare
    cdef object _result
    cdef object _target

    def __cinit__(self):
        self.oaddr  = NULL
        self.ocount = 0
        self.otype  = MPI_DATATYPE_NULL
        self.raddr  = NULL
        self.rcount = 0
        self.rtype  = MPI_DATATYPE_NULL
        self.tdisp  = 0
        self.tcount = 0
        self.ttype  = MPI_DATATYPE_NULL

    cdef int for_rma(self, int readonly,
                     object origin, int rank, object target) except -1:
        # ORIGIN
        self._origin = message_simple(
            origin, readonly, rank, 0,
            &self.oaddr,  &self.ocount,  &self.otype)
        if (
            (rank == MPI_PROC_NULL) and
            (origin is not None) and
            (is_list(origin) or is_tuple(origin)) and
            (len(origin) > 0 and is_datatype(origin[-1]))
        ):
            self.otype  = asdatatype(origin[-1]).ob_mpi
            self._origin = origin
        # TARGET
        cdef Py_ssize_t nargs = 0
        if target is None:
            self.tdisp  = 0
            self.tcount = self.ocount
            self.ttype  = self.otype
        elif is_integral(target):
            self.tdisp  = <MPI_Aint> PyNumber_Index(target)
            self.tcount = self.ocount
            self.ttype  = self.otype
        elif is_list(target) or is_tuple(target):
            self.tdisp  = 0
            self.tcount = self.ocount
            self.ttype  = self.otype
            nargs = len(target)
            if nargs >= 1:
                self.tdisp  = <MPI_Aint> PyNumber_Index(target[0])
            if nargs >= 2:
                self.tcount = <MPI_Count> PyNumber_Index(target[1])
            if nargs >= 3:
                self.ttype  = asdatatype(target[2]).ob_mpi
            if nargs >= 4:
                raise ValueError("target: expecting 3 items at most")
        else:
            raise ValueError("target: expecting integral or list/tuple")
        self._target = target
        return 0

    cdef int for_put(self, object origin, int rank, object target) except -1:
        self.for_rma(1, origin, rank, target)
        return 0

    cdef int for_get(self, object origin, int rank, object target) except -1:
        self.for_rma(0, origin, rank, target)
        return 0

    cdef int for_acc(self, object origin, int rank, object target) except -1:
        self.for_rma(1, origin, rank, target)
        return 0

    cdef int set_origin(self, object origin, int rank) except -1:
        self._origin = message_simple(
            origin, 1, rank, 0,
            &self.oaddr,  &self.ocount,  &self.otype)
        self.tdisp  = 0
        self.tcount = self.ocount
        self.ttype  = self.otype

    cdef int set_compare(self, object compare, int rank) except -1:
        self._compare = message_simple(
            compare, 1, rank, 0,
            &self.caddr,  &self.ccount,  &self.ctype)

    cdef int set_result(self, object result, int rank) except -1:
        self._result = message_simple(
            result, 0, rank, 0,
            &self.raddr,  &self.rcount,  &self.rtype)

    cdef int for_get_acc(self, object origin, object result,
                         int rank, object target) except -1:
        self.for_rma(0, origin, rank, target)
        self.set_result(result, rank)
        return 0

    cdef int for_fetch_op(self, object origin, object result,
                          int rank, MPI_Aint disp) except -1:
        self.set_origin(origin, rank)
        self.set_result(result, rank)
        self.tdisp = disp
        if rank == MPI_PROC_NULL: return 0
        # Check
        if self.ocount != 1:
            raise ValueError(
                f"origin: expecting a single element, got {self.ocount}")
        if self.rcount != 1:
            raise ValueError(
                f"result: expecting a single element, got {self.rcount}")
        if self.otype != self.rtype:
            raise ValueError(
                f"mismatch in origin and result MPI datatypes")
        return 0

    cdef int for_cmp_swap(self, object origin, object compare, object result,
                          int rank, MPI_Aint disp) except -1:
        self.set_origin(origin, rank)
        self.set_compare(compare, rank)
        self.set_result(result, rank)
        self.tdisp = disp
        if rank == MPI_PROC_NULL: return 0
        # Check
        if self.ocount != 1:
            raise ValueError(
                f"origin: expecting a single element, got {self.ocount}")
        if self.ccount != 1:
            raise ValueError(
                f"compare: expecting a single element, got {self.ccount}")
        if self.rcount != 1:
            raise ValueError(
                f"result: expecting a single element, got {self.rcount}")
        if self.otype != self.ctype:
            raise ValueError(
                f"mismatch in origin and compare MPI datatypes")
        if self.otype != self.rtype:
            raise ValueError(
                f"mismatch in origin and result MPI datatypes")
        return 0


cdef inline _p_msg_rma message_rma():
    cdef _p_msg_rma msg = _p_msg_rma.__new__(_p_msg_rma)
    return msg

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_msg_io:

    # raw C-side data
    cdef void         *buf
    cdef MPI_Count    count
    cdef MPI_Datatype dtype
    # python-side data
    cdef object _msg

    def __cinit__(self):
        self.buf   = NULL
        self.count = 0
        self.dtype = MPI_DATATYPE_NULL

    cdef int for_read(self, object msg) except -1:
        self._msg = message_simple(
            msg, 0,  # writable
            0, 0,
            &self.buf,
            &self.count,
            &self.dtype,
        )
        return 0

    cdef int for_write(self, object msg) except -1:
        self._msg = message_simple(
            msg, 1,  # readonly
            0, 0,
            &self.buf,
            &self.count,
            &self.dtype,
        )
        return 0


cdef inline _p_msg_io message_io_read(object buf):
    cdef _p_msg_io msg = _p_msg_io.__new__(_p_msg_io)
    msg.for_read(buf)
    return msg

cdef inline _p_msg_io message_io_write(object buf):
    cdef _p_msg_io msg = _p_msg_io.__new__(_p_msg_io)
    msg.for_write(buf)
    return msg

# -----------------------------------------------------------------------------



import os
import sys

from mpi4py import rc

vmap = {
    "true": True,
    "false": False,
    "0": 0,
    "1": 1,
    "1024": 1024,
}
for arg in sys.argv[1:]:
    attr, value = arg.split("=")
    setattr(rc, attr, vmap.get(value, value))

if rc.errors == "abort":
    rc.initialize = False

from mpi4py import MPI  # noqa: E402

if rc.errors == "abort":
    vendor, version = MPI.get_vendor()
    if vendor == "Intel MPI":
        if version[:2] < (2021, 11):
            sys.exit(0)
    if vendor == "MPICH":
        if version[:2] < (4, 1):
            sys.exit(0)
    if vendor == "Open MPI":
        if version[:2] < (5, 0):
            sys.exit(0)


def check_errhandler(obj):
    if rc.errors == "default":
        if isinstance(obj, MPI.File):
            check_eh = MPI.ERRORS_RETURN
        else:
            check_eh = MPI.ERRORS_ARE_FATAL
    elif rc.errors == "exception":
        check_eh = MPI.ERRORS_RETURN
    elif rc.errors == "abort":
        check_eh = MPI.ERRORS_ABORT
        if MPI.ERRORS_ABORT == MPI.ERRHANDLER_NULL:
            check_eh = MPI.ERRORS_ARE_FATAL
    elif rc.errors == "fatal":
        check_eh = MPI.ERRORS_ARE_FATAL
    else:
        assert 0
    eh = obj.Get_errhandler()
    try:
        assert eh == check_eh
    finally:
        if eh != MPI.ERRHANDLER_NULL:
            eh.Free()


if not MPI.Is_initialized():
    MPI.Init()


try:
    session = MPI.Session.Init()
    try:
        check_errhandler(session)
    finally:
        session.Finalize()
except NotImplementedError:
    pass
except MPI.Exception:
    pass


try:
    for commbase in (MPI.COMM_SELF, MPI.COMM_WORLD):
        check_errhandler(commbase)
        comm = commbase.Dup()
        try:
            check_errhandler(comm)
        finally:
            comm.Free()
except NotImplementedError:
    pass
except MPI.Exception:
    pass


weh = MPI.COMM_SELF.Get_errhandler()
MPI.COMM_SELF.Set_errhandler(MPI.ERRORS_RETURN)
try:
    win = MPI.Win.Create(
        MPI.BOTTOM,
        1,
        MPI.INFO_NULL,
        MPI.COMM_SELF,
    )
    try:
        check_errhandler(win)
    finally:
        win.Free()
except NotImplementedError:
    pass
except MPI.Exception:
    pass
finally:
    MPI.COMM_SELF.Set_errhandler(weh)
    weh.Free()


try:
    # check_errhandler(MPI.FILE_NULL)  # TODO
    file = MPI.File.Open(
        MPI.COMM_SELF,
        os.devnull,
        MPI.MODE_WRONLY,
        MPI.INFO_NULL,
    )
    try:
        check_errhandler(file)
    finally:
        file.Close()
except NotImplementedError:
    pass
except MPI.Exception:
    pass


if not MPI.Is_finalized():
    MPI.Finalize()
